from dataclasses import asdict
from typing import Any, Dict, Optional

import torch
from sarathi.config import ParallelConfig

from vidur.config.model_config import BaseModelConfig
from vidur.types import ActivationType, NormType


class ModelConfig:
    def __init__(
        self,
        name: str,
        num_layers: int,
        num_q_heads: int,
        num_kv_heads: int,
        embedding_dim: int,
        mlp_hidden_dim: int,
        max_model_len: int,
        use_gated_mlp: bool,
        use_bias: bool,
        use_qkv_bias: bool,
        activation: ActivationType,
        norm: NormType,
        post_attn_norm: bool,
        vocab_size: int,
        is_neox_style: Optional[bool] = True,
        rope_theta: Optional[int] = None,
        rope_scaling: Optional[Dict[str, Any]] = None,
        partial_rotary_factor: float = 1.0,
        no_tensor_parallel: bool = False,
    ):
        self.name = name
        self.num_layers = num_layers
        self.num_q_heads = num_q_heads
        self.num_kv_heads = num_kv_heads
        self.embedding_dim = embedding_dim
        self.mlp_hidden_dim = mlp_hidden_dim
        self.max_model_len = max_model_len
        self.use_gated_mlp = use_gated_mlp
        self.vocab_size = vocab_size
        self.use_bias = use_bias
        self.use_qkv_bias = use_qkv_bias
        self.activation = str(activation)
        self.norm = str(norm)
        self.post_attn_norm = post_attn_norm
        self.no_tensor_parallel = no_tensor_parallel
        self.partial_rotary_factor = partial_rotary_factor
        self.rope_theta = rope_theta
        self.rope_scaling = rope_scaling
        self.is_neox_style = is_neox_style

        assert self.norm in ["layer_norm", "rms_norm"]
        assert self.activation in ["gelu", "silu"]

        if self.use_gated_mlp:
            assert self.activation == "silu"
        else:
            assert self.activation == "gelu"

    @staticmethod
    def from_model_name(model_name: str, max_model_len: int = 4096):
        model_config: BaseModelConfig = BaseModelConfig.create_from_name(model_name)
        model_config_dict = asdict(model_config)

        model_config = ModelConfig(model_name, **model_config_dict)
        model_config.max_model_len = max_model_len

        return model_config

    def get_num_q_heads(self, parallel_config: ParallelConfig):
        return self.num_q_heads // parallel_config.tensor_parallel_size

    def get_num_kv_heads(self, parallel_config: ParallelConfig):
        return self.num_kv_heads // parallel_config.tensor_parallel_size

    def get_head_size(self):
        return self.embedding_dim // self.num_q_heads

    @property
    def dtype(self):
        return torch.float16



import argparse
import os

import pandas as pd
import streamlit as st

st.set_page_config(page_title="Vidur Demo", layout="wide")

from vidur.config_optimizer.analyzer.constants import AXIS_COLS, PRETTY_NAMES
from vidur.config_optimizer.analyzer.dashboard.best_config_page import (
    render_best_config_selection_page,
)
from vidur.config_optimizer.analyzer.dashboard.config_compare_page import (
    render_config_comparison_page,
)
from vidur.config_optimizer.analyzer.dashboard.cost_analysis_page import (
    render_cost_analysis_page,
)
from vidur.config_optimizer.analyzer.dashboard.intro_page import render_intro_page
from vidur.config_optimizer.analyzer.dashboard.pareto_curve_page import (
    render_pareto_curve_page,
)
from vidur.config_optimizer.analyzer.dashboard.search_analysis_page import (
    render_search_analysis_page,
)
from vidur.config_optimizer.analyzer.dashboard.utils import (
    get_config_name,
    hide_anchor_link,
)
from vidur.config_optimizer.analyzer.utils import get_trace_name


def select_best_run(
    df: pd.DataFrame,
    scheduling_delay_slo_percentile: float,
    scheduling_delay_slo_value: float,
):
    metric_names = f"request_scheduling_delay_{scheduling_delay_slo_percentile}%"
    mask = df[metric_names] < scheduling_delay_slo_value
    if len(df[mask]) > 0:
        # multiple points satisfy the SLO
        # select the point with the highest qps
        return (
            df[mask]
            .sort_values("poisson_request_interval_generator_qps", ascending=False)
            .iloc[0]
        )
    else:
        # no point satisfy the SLO
        # select the point with the lowest scheduling delay
        return df.iloc[df[metric_names].argmin()]


@st.cache_data
def prepare_subset_dfs(_args):
    analysis_dir = f"{_args.sim_results_dir}/analysis"

    stats_file = f"{analysis_dir}/stats.csv"

    if not os.path.exists(stats_file):
        raise ValueError(f"Stats file {stats_file} does not exist")

    df = pd.read_csv(stats_file)

    df = df[df["replica_device"] != "a40"]

    OLD_GPU_COSTS = {
        "h100": 4.25,
        "a100": 2.21,
        "a40": 1.28,
    }

    NEW_GPU_COSTS = {
        "h100": 6.98,
        "a100": 3.72,
        "a40": 1.28,
    }

    # update the "cost" and "capacity_per_dollar" columns accourding to the new costs use "replica_device" as the key
    def update_cost(row):
        if row["replica_device"] == "a40":
            return row["cost"]
        return (
            NEW_GPU_COSTS[row["replica_device"]]
            * row["cost"]
            / OLD_GPU_COSTS[row["replica_device"]]
        )

    def update_capacity_per_dollar(row):
        if row["replica_device"] == "a40":
            return row["capacity_per_dollar"]
        return (
            row["capacity_per_dollar"]
            * OLD_GPU_COSTS[row["replica_device"]]
            / NEW_GPU_COSTS[row["replica_device"]]
        )

    df["cost"] = df.apply(update_cost, axis=1)
    df["capacity_per_dollar"] = df.apply(update_capacity_per_dollar, axis=1)

    # map the trace file to a human readable name
    df["Trace"] = df["trace_request_length_generator_trace_file"].apply(get_trace_name)
    # map axis columns with human readable names
    df = df.rename(columns=AXIS_COLS)
    # replace scheduler and SKU names with human readable names
    df["Scheduler"] = df["Scheduler"].replace(PRETTY_NAMES)
    df["SKU"] = df["SKU"].replace(PRETTY_NAMES)

    df["Model"] = df["replica_model_name"].str.split("/").str[1]
    df["Model"] = df["Model"].str.replace("-Instruct-hf", "")
    df["Model"] = df["Model"].str.replace("-hf", "")

    # rename all cols starting with tbt_ and ttft_ to TBT_ and TTFT_
    df = df.rename(
        columns=lambda x: x.replace("tbt_", "TBT_").replace("ttft_", "TTFT_")
    )
    # create str version of capacity_per_dollar
    df["capacity_per_dollar_str"] = df["capacity_per_dollar"].apply(
        lambda x: str(round(x, 3))
    )
    # rename capacity_per_dollar to QPS per Dollar
    df = df.rename(columns={"capacity_per_dollar": "QPS per Dollar"})
    # convert TBT_ cols to ms from seconds
    tbt_cols = [col for col in df.columns if "TBT_" in col]
    # remove tbt_cdf
    tbt_cols.remove("TBT_cdf")
    df[tbt_cols] = df[tbt_cols] * 1000

    # if scheduler is not sarathi then set sarathi chunk size to -
    df.loc[df["Scheduler"] != "Sarathi-Serve", "Sarathi Chunk Size"] = "-"

    # assign a unique name to each config
    df["Name"] = df.apply(get_config_name, axis=1)

    filtered_subset_dfs = {}

    for (model_name, trace_name), group in df.groupby(["Model", "Trace"]):
        filtered_subset_df = (
            group.groupby(list(AXIS_COLS.values()))
            .apply(
                lambda x: select_best_run(
                    x,
                    _args.scheduling_delay_slo_percentile,
                    _args.scheduling_delay_slo_value,
                )
            )
            .reset_index(drop=True)
        )
        filtered_subset_dfs[(model_name, trace_name)] = filtered_subset_df

    # combine all the filtered subset dfs
    filtered_df = pd.concat(filtered_subset_dfs.values())

    return df, filtered_df, filtered_subset_dfs


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--sim-results-dir", type=str, required=True)
    parser.add_argument("--scheduling-delay-slo-percentile", type=float, default=95)
    parser.add_argument("--scheduling-delay-slo-value", type=float, default=2.0)
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()

    df, filtered_df, filtered_subset_dfs = prepare_subset_dfs(args)

    hide_anchor_link()

    if "comparison_points" not in st.session_state:
        st.session_state.comparison_points = []

    page_names_to_funcs = {
        "-": render_intro_page,
        "Best Config Selection": lambda: render_best_config_selection_page(filtered_df),
        "Config Comparison": lambda: render_config_comparison_page(filtered_df),
        "Pareto Curve Analysis": lambda: render_pareto_curve_page(filtered_subset_dfs),
        "Cost Analysis": lambda: render_cost_analysis_page(filtered_subset_dfs),
    }

    if args.debug:
        page_names_to_funcs["Search Analysis"] = lambda: render_search_analysis_page(
            filtered_df
        )

    demo_name = st.sidebar.selectbox("Choose analysis mode", page_names_to_funcs.keys())
    page_names_to_funcs[demo_name]()


if __name__ == "__main__":
    main()



from vidur.config import ReplicaConfig
from vidur.entities import BatchStage
from vidur.utils.param_counter import ParamCounter


class MFUCalculator:

    def __init__(self, replica_config: ReplicaConfig):
        param_counter = ParamCounter(replica_config)
        self._num_params_per_device = param_counter.get_num_parameters_per_device()

        model_config = replica_config.model_config

        self._num_layers_per_device = (
            model_config.num_layers // replica_config.num_pipeline_stages
        )
        self._num_heads_per_device = (
            model_config.num_q_heads // replica_config.tensor_parallel_size
        )
        self._head_dimension = model_config.embedding_dim // model_config.num_q_heads
        self._device_flops = replica_config.device_config.fp16_tflops * 2**40

    def _get_mlp_flops(self, batch_stage: BatchStage) -> float:
        num_tokens = sum(batch_stage.num_q_tokens)
        return 2 * num_tokens * self._num_params_per_device

    def _get_attention_flops(self, batch_stage: BatchStage) -> float:
        total_flops = 0
        for request, num_q_tokens, num_kv_tokens in zip(
            batch_stage.requests, batch_stage.num_q_tokens, batch_stage.num_kv_tokens
        ):
            total_flops += (
                4  # for number of ops in attention
                * self._num_layers_per_device
                * self._num_heads_per_device
                * self._head_dimension
                * num_q_tokens
                * num_kv_tokens
            )

        return total_flops

    def get_mfu(self, batch_stage: BatchStage) -> float:
        mlp_flops = self._get_mlp_flops(batch_stage)
        attention_flops = self._get_attention_flops(batch_stage)
        total_flops = mlp_flops + attention_flops
        total_flops_per_second = total_flops / batch_stage.execution_time
        return total_flops_per_second * 100 / self._device_flops



import numpy as np

EPS = 1e-8


class ZipfGenerator:
    def __init__(
        self, min: int, max: int, theta: float, scramble: bool, seed: int
    ) -> None:
        self._min = min
        self._max = max
        self._items = max - min + 1
        self._theta = theta
        self._zeta_2 = self._zeta(2, self._theta)
        self._alpha = 1.0 / (1.0 - self._theta)
        self._zetan = self._zeta(self._items, self._theta)
        self._eta = (1 - np.power(2.0 / self._items, 1 - self._theta)) / (
            1 - self._zeta_2 / (self._zetan + EPS)
        )
        self._scramble = scramble
        self._seed = seed
        self._generator = np.random.RandomState(seed)

    def _zeta(self, count: float, theta: float) -> float:
        return np.sum(1 / (np.power(np.arange(1, count), theta)))

    def _next(self) -> int:
        u = self._generator.random_sample()
        uz = u * self._zetan

        if uz < 1.0:
            return self._min

        if uz < 1.0 + np.power(0.5, self._theta):
            return self._min + 1

        return self._min + int(
            (self._items) * np.power(self._eta * u - self._eta + 1, self._alpha)
        )

    def next(self) -> int:
        retval = self._next()
        if self._scramble:
            retval = self._min + hash(str(retval) + str(self._seed)) % self._items

        return retval



from math import ceil

from vidur.config import ReplicaConfig


class ParamCounter:
    def __init__(self, replica_config: ReplicaConfig) -> None:
        self._replica_config = replica_config
        self._model_config = self._replica_config.model_config

        assert (
            self._model_config.num_q_heads % self._replica_config.tensor_parallel_size
            == 0
        )
        assert (
            self._model_config.num_layers % self._replica_config.num_pipeline_stages
            == 0
        )
        assert (
            self._model_config.embedding_dim % self._replica_config.tensor_parallel_size
            == 0
        )
        assert self._model_config.embedding_dim % self._model_config.num_q_heads == 0

        self._num_layers_per_pipeline_stage = (
            self._model_config.num_layers // self._replica_config.num_pipeline_stages
        )
        self._attention_head_dim = (
            self._model_config.embedding_dim // self._model_config.num_q_heads
        )
        self._q_heads_per_tensor_parallel_worker = (
            self._model_config.num_q_heads // self._replica_config.tensor_parallel_size
        )
        self._kv_heads_per_tensor_parallel_worker = ceil(
            self._model_config.num_kv_heads / self._replica_config.tensor_parallel_size
        )

    def get_num_parameters_per_layer(self) -> int:
        num_parameters = 0
        # weights for attention metrics Wq, Wk, Wv
        num_parameters += (
            self._model_config.embedding_dim
            * self._attention_head_dim
            * (
                self._q_heads_per_tensor_parallel_worker
                + 2 * self._kv_heads_per_tensor_parallel_worker
            )
        )
        # weights for attention metrics Wo
        num_parameters += (
            self._model_config.embedding_dim
            * self._attention_head_dim
            * self._q_heads_per_tensor_parallel_worker
        )
        # fc layer weights
        if self._model_config.use_gated_mlp:
            num_parameters += (
                3
                * self._model_config.embedding_dim
                * self._model_config.mlp_hidden_dim
                // self._replica_config.tensor_parallel_size
            )
        else:
            num_parameters += (
                2
                * self._model_config.embedding_dim
                * self._model_config.mlp_hidden_dim
                // self._replica_config.tensor_parallel_size
            )

        return num_parameters

    def get_num_parameters_per_device(self) -> int:
        num_parameters_per_layer = self.get_num_parameters_per_layer()
        return num_parameters_per_layer * self._num_layers_per_pipeline_stage



from math import ceil
from typing import List

from vidur.entities.batch import Batch, Request
from vidur.scheduler.replica_scheduler.base_replica_scheduler import (
    BaseReplicaScheduler,
)


class VLLMReplicaScheduler(BaseReplicaScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # For vLLM and its derivatives, we only need to set a loose max batch size
        # Memory requirements are handled explicitly by the scheduler
        self._max_micro_batch_size = self._config.batch_size_cap // self._num_stages
        self._watermark_blocks = int(
            self._config.watermark_blocks_fraction * self._config.num_blocks
        )

    def on_batch_end(self, batch: Batch) -> None:
        super().on_batch_end(batch)

        for request in batch.requests:
            if request.completed:
                self.free(request.id)
            else:
                self._preempted_requests.append(request)

    def _can_allocate_request(self, request: Request) -> bool:
        if request.id not in self._allocation_map:
            # new request
            num_required_blocks = ceil(
                (request.num_prefill_tokens) / self._replica_config.block_size
            )
            return (
                self._config.num_blocks
                - self._num_allocated_blocks
                - num_required_blocks
                >= self._watermark_blocks
            )

        # vllm requires at least one block to be available
        return self._config.num_blocks - self._num_allocated_blocks >= 1

    def _allocate_request(self, request: Request) -> None:
        if request.id not in self._allocation_map:
            # new request
            num_required_blocks = ceil(
                (request.num_prefill_tokens) / self._replica_config.block_size
            )
            self.allocate(request.id, num_required_blocks)
            return

        num_tokens_reserved = (
            self._allocation_map[request.id] * self._replica_config.block_size
        )
        num_tokens_required = max(0, request.num_processed_tokens - num_tokens_reserved)
        assert (
            num_tokens_required == 0 or num_tokens_required == 1
        ), f"num_tokens_required: {num_tokens_required}"

        if num_tokens_required == 0:
            return

        self.allocate(request.id, 1)

    def _get_next_batch(self, time: float) -> Batch:
        requests = []
        num_q_tokens = []
        num_kv_tokens = []
        total_num_q_tokens = 0

        while self._request_queue:
            request = self._request_queue[0]

            next_num_q_tokens = self._get_request_next_num_q_tokens(request)
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens

            if not self._can_allocate_request(request):
                break

            if (
                total_num_q_tokens + next_num_q_tokens
                > self._config.max_tokens_in_batch
            ):
                break

            if len(self._allocation_map) == self._config.batch_size_cap:
                break

            if len(requests) == self._max_micro_batch_size:
                break

            request = self._request_queue.pop(0)

            self._allocate_request(request)
            requests.append(request)
            num_q_tokens.append(next_num_q_tokens)
            num_kv_tokens.append(next_num_kv_tokens)
            total_num_q_tokens += next_num_q_tokens

        if requests:
            return Batch(self._replica_id, requests, num_q_tokens, num_kv_tokens)

        # Safer to sort preempted_requests to maintain FIFO order
        self._preempted_requests.sort(key=lambda r: r.arrived_at)
        # all preempted_requests will have prefill completed
        while self._preempted_requests:
            if len(requests) == self._max_micro_batch_size:
                break

            request = self._preempted_requests.pop(0)

            while not self._can_allocate_request(request):
                if self._preempted_requests:
                    victim_request = self._preempted_requests.pop(-1)
                    victim_request.restart()
                    self.free(victim_request.id)
                    self._request_queue = [victim_request] + self._request_queue
                else:
                    request.restart()
                    self.free(request.id)
                    self._request_queue = [request] + self._request_queue
                    break
            else:
                self._allocate_request(request)
                next_num_q_tokens = self._get_request_next_num_q_tokens(request)
                next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens
                requests.append(request)
                num_q_tokens.append(next_num_q_tokens)
                num_kv_tokens.append(next_num_kv_tokens)

        if not requests:
            return

        return Batch(self._replica_id, requests, num_q_tokens, num_kv_tokens)



from vidur.types.activation_type import ActivationType
from vidur.types.base_int_enum import BaseIntEnum
from vidur.types.device_sku_type import DeviceSKUType
from vidur.types.event_type import EventType
from vidur.types.execution_time_predictor_type import ExecutionTimePredictorType
from vidur.types.global_scheduler_type import GlobalSchedulerType
from vidur.types.node_sku_type import NodeSKUType
from vidur.types.norm_type import NormType
from vidur.types.replica_scheduler_type import ReplicaSchedulerType
from vidur.types.request_generator_type import RequestGeneratorType
from vidur.types.request_interval_generator_type import RequestIntervalGeneratorType
from vidur.types.request_length_generator_type import RequestLengthGeneratorType

__all__ = [
    EventType,
    ExecutionTimePredictorType,
    GlobalSchedulerType,
    RequestGeneratorType,
    RequestLengthGeneratorType,
    RequestIntervalGeneratorType,
    ReplicaSchedulerType,
    DeviceSKUType,
    NodeSKUType,
    NormType,
    ActivationType,
    BaseIntEnum,
]



from abc import ABC, abstractmethod
from typing import List

from vidur.metrics import MetricsStore
from vidur.scheduler import BaseGlobalScheduler
from vidur.types import EventType


class BaseEvent(ABC):
    _id = 0

    def __init__(self, time: float, event_type: EventType):
        self._time = time
        self._id = BaseEvent.generate_id()
        self._event_type = event_type
        self._priority_number = self._get_priority_number()

    @classmethod
    def generate_id(cls):
        cls._id += 1
        return cls._id

    @property
    def id(self) -> int:
        return self._id

    @property
    def time(self):
        return self._time

    @property
    def event_type(self):
        pass

    @abstractmethod
    def handle_event(
        self,
        current_time: float,
        scheduler: BaseGlobalScheduler,
        metrics_store: MetricsStore,
    ) -> List["BaseEvent"]:
        pass

    def _get_priority_number(self):
        return (self._time, self._id, self.event_type)

    def __lt__(self, other):
        if self._time == other._time:
            if self._event_type == other._event_type:
                return self._id < other._id
            return self._event_type < other._event_type
        else:
            return self._time < other._time

    def __eq__(self, other):
        return (
            self._time == other._time
            and self._event_type == other._event_type
            and self._id == other._id
        )

    def __str__(self) -> str:
        # use to_dict to get a dict representation of the object
        # and convert it to a string
        class_name = self.__class__.__name__
        return f"{class_name}({str(self.to_dict())})"

    def to_dict(self):
        return {"time": self.time, "event_type": self.event_type}

    def to_chrome_trace_events(self) -> dict:
        return None



from random import randint
from typing import List, Tuple

from vidur.entities import Request
from vidur.scheduler.global_scheduler.base_global_scheduler import BaseGlobalScheduler


class RandomGlobalScheduler(BaseGlobalScheduler):
    def schedule(self) -> List[Tuple[int, Request]]:
        self.sort_requests()

        request_mapping = []
        while self._request_queue:
            request = self._request_queue.pop(0)
            replica_id = randint(1, self._num_replicas) - 1
            request_mapping.append((replica_id, request))
        return request_mapping



from vidur.scheduler.replica_scheduler.faster_transformer_replica_scheduler import (
    FasterTransformerReplicaScheduler,
)
from vidur.scheduler.replica_scheduler.lightllm_replica_scheduler import (
    LightLLMReplicaScheduler,
)
from vidur.scheduler.replica_scheduler.mnemosyne_edf_replica_scheduler import (
    MnemosyneEDFReplicaScheduler,
)
from vidur.scheduler.replica_scheduler.mnemosyne_fcfs_fixed_chunk_replica_scheduler import (
    MnemosyneFCFSFixedChunkReplicaScheduler,
)
from vidur.scheduler.replica_scheduler.mnemosyne_fcfs_replica_scheduler import (
    MnemosyneFCFSReplicaScheduler,
)
from vidur.scheduler.replica_scheduler.mnemosyne_lrs_replica_scheduler import (
    MnemosyneLRSReplicaScheduler,
)
from vidur.scheduler.replica_scheduler.mnemosyne_st_replica_scheduler import (
    MnemosyneSTReplicaScheduler,
)
from vidur.scheduler.replica_scheduler.orca_replica_scheduler import (
    OrcaReplicaScheduler,
)
from vidur.scheduler.replica_scheduler.sarathi_replica_scheduler import (
    SarathiReplicaScheduler,
)
from vidur.scheduler.replica_scheduler.vllm_replica_scheduler import (
    VLLMReplicaScheduler,
)
from vidur.types import ReplicaSchedulerType
from vidur.utils.base_registry import BaseRegistry


class ReplicaSchedulerRegistry(BaseRegistry):
    pass


ReplicaSchedulerRegistry.register(
    ReplicaSchedulerType.FASTER_TRANSFORMER, FasterTransformerReplicaScheduler
)
ReplicaSchedulerRegistry.register(ReplicaSchedulerType.ORCA, OrcaReplicaScheduler)
ReplicaSchedulerRegistry.register(ReplicaSchedulerType.SARATHI, SarathiReplicaScheduler)
ReplicaSchedulerRegistry.register(ReplicaSchedulerType.VLLM, VLLMReplicaScheduler)
ReplicaSchedulerRegistry.register(
    ReplicaSchedulerType.LIGHTLLM, LightLLMReplicaScheduler
)
ReplicaSchedulerRegistry.register(
    ReplicaSchedulerType.MNEMOSYNE_FCFS_FIXED_CHUNK,
    MnemosyneFCFSFixedChunkReplicaScheduler,
)
ReplicaSchedulerRegistry.register(
    ReplicaSchedulerType.MNEMOSYNE_FCFS, MnemosyneFCFSReplicaScheduler
)
ReplicaSchedulerRegistry.register(
    ReplicaSchedulerType.MNEMOSYNE_EDF, MnemosyneEDFReplicaScheduler
)
ReplicaSchedulerRegistry.register(
    ReplicaSchedulerType.MNEMOSYNE_LRS, MnemosyneLRSReplicaScheduler
)
ReplicaSchedulerRegistry.register(
    ReplicaSchedulerType.MNEMOSYNE_ST, MnemosyneSTReplicaScheduler
)



from abc import ABC
from dataclasses import dataclass
from typing import Any

from vidur.config.utils import get_all_subclasses


@dataclass
class BasePolyConfig(ABC):

    @classmethod
    def create_from_type(cls, type_: Any) -> Any:
        for subclass in get_all_subclasses(cls):
            if subclass.get_type() == type_:
                return subclass()
        raise ValueError(f"Invalid type: {type_}")



import numpy as np
import torch

from vidur.profiling.collectives.collectives_impl import GraphedCollective
from vidur.profiling.common.cuda_timer import CudaTimer
from vidur.profiling.common.timer_stats_store import TimerStatsStore

WARMUP_STEPS = 1
ACTIVE_STEPS = 3
GRAPH_DISABLED_STEPS = 10
DISABLE_GRAPH = True


class CollectiveWrapper:
    def __init__(
        self,
        rank: int,
        num_workers: int,
        comm_id: int,
        size: int,
        collective: str,
        devices_per_node: int,
        max_devices_per_node: int,
    ) -> None:
        self._rank = rank
        self._num_workers = num_workers
        self._size = size
        self._comm_id = comm_id
        self._collective = collective
        self._devices_per_node = devices_per_node
        self._max_devices_per_node = max_devices_per_node

        self._graphed_collective = GraphedCollective(
            num_workers, size, collective=collective, disable_graph=DISABLE_GRAPH
        )

        self.timer_stats_store = TimerStatsStore(profile_method="kineto")
        self._cuda_timer = CudaTimer(
            collective, aggregation_fn=np.median, filter_str="nccl"
        )

    def _run_collective(self):
        torch.cuda.synchronize()
        torch.distributed.barrier()

        with self._cuda_timer:
            if DISABLE_GRAPH:
                for _ in range(GRAPH_DISABLED_STEPS):
                    self._graphed_collective.launch()

            self._graphed_collective.launch()

        torch.cuda.synchronize()

    def profile(self):
        self.timer_stats_store.clear_stats()
        for _ in range(ACTIVE_STEPS):
            self._run_collective()

        return {
            "time_stats": self.timer_stats_store.get_stats(),
            "rank": self._rank,
            "num_workers": self._num_workers,
            "size": self._size * 2,  # bytes
            "collective": self._collective,
            "devices_per_node": self._devices_per_node,
            "max_devices_per_node": self._max_devices_per_node,
        }



from vidur.execution_time_predictor.base_execution_time_predictor import (
    BaseExecutionTimePredictor,
)
from vidur.execution_time_predictor.execution_time_predictor_registry import (
    ExecutionTimePredictorRegistry,
)

__all__ = [ExecutionTimePredictorRegistry, BaseExecutionTimePredictor]



from typing import List

from vidur.entities import Request
from vidur.events.base_event import BaseEvent
from vidur.logger import init_logger
from vidur.metrics import MetricsStore
from vidur.scheduler import BaseGlobalScheduler
from vidur.types import EventType

logger = init_logger(__name__)


class RequestArrivalEvent(BaseEvent):
    def __init__(self, time: float, request: Request) -> None:
        super().__init__(time, EventType.REQUEST_ARRIVAL)

        self._request = request

    def handle_event(
        self, scheduler: BaseGlobalScheduler, metrics_store: MetricsStore
    ) -> List[BaseEvent]:
        from vidur.events.global_schedule_event import GlobalScheduleEvent

        logger.debug(f"Request: {self._request.id} arrived at {self.time}")
        scheduler.add_request(self._request)
        metrics_store.on_request_arrival(self.time, self._request)
        return [GlobalScheduleEvent(self.time)]

    def to_dict(self) -> dict:
        return {
            "time": self.time,
            "event_type": self.event_type,
            "request": self._request.id,
        }



import argparse
import datetime
import os
from typing import Any, List

import pandas as pd
import ray
import torch
from sarathi.config import ParallelConfig
from sarathi.model_executor.attention import AttentionBackend
from tqdm import tqdm

from vidur.profiling.attention.attention_input import AttentionInput
from vidur.profiling.attention.attention_wrapper import AttentionWrapper
from vidur.profiling.common.model_config import ModelConfig
from vidur.profiling.utils import get_attention_input_combinations, get_max_num_blocks


def parse_args():
    parser = argparse.ArgumentParser(description="Attention Profiling")
    parser.add_argument(
        "--disable_ray",
        action="store_true",
        help="Disable Ray",
    )
    parser.add_argument(
        "--num_gpus",
        type=int,
        default=8,
        help="Number of GPUs to use for profiling",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="profiling_outputs",
        help="Output directory for profiling results",
    )
    parser.add_argument(
        "--models",
        type=str,
        nargs="+",
        default=[
            "meta-llama/Llama-3-8B",
            "meta-llama/Meta-Llama-3-70B",
        ],
        help="Models to profile",
    )
    parser.add_argument(
        "--num_tensor_parallel_workers",
        type=int,
        nargs="+",
        default=[1, 2, 4, 8],
        help="Number of tensor parallel workers to profile",
    )
    parser.add_argument(
        "--max_seq_len",
        type=int,
        default=8 * 1024 * 1024,
        help="Maximum context length of input",
    )
    parser.add_argument(
        "--min_batch_size",
        type=int,
        default=1,
        help="Maximum decode batch size",
    )
    parser.add_argument(
        "--max_batch_size",
        type=int,
        default=1,
        help="Maximum decode batch size",
    )
    parser.add_argument(
        "--profile_only_decode",
        action="store_true",
        help="Only profile the decode",
    )
    parser.add_argument(
        "--profile_only_prefill",
        action="store_true",
        help="Only profile the prefill",
    )
    parser.add_argument(
        "--attention_backend",
        default=AttentionBackend.FLASHINFER,
        choices=[e.value for e in AttentionBackend],
        help="The attention backend to profile (default: %(default)s)",
    )
    parser.add_argument(
        "--block_size",
        type=int,
        default=16,
        help="Block size for paged attention",
    )
    parser.add_argument(
        "--max_chunk_size",
        type=int,
        default=4096,
        help="Maximum chunk size for chunked prefill",
    )
    args = parser.parse_args()

    args.output_dir = f"{args.output_dir}/attention/{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"
    os.makedirs(args.output_dir, exist_ok=True)

    return args


def profile_model(
    args: argparse.Namespace,
    model: str,
    num_tensor_parallel_workers: int,
    input_combinations: List[AttentionInput],
    max_num_blocks: int,
    dtype: torch.dtype,
    pbar: Any,
):
    model_config = ModelConfig.from_model_name(model)
    parallel_config = ParallelConfig(
        tensor_parallel_size=num_tensor_parallel_workers,
        pipeline_parallel_size=1,
    )

    promises = []
    all_results = []

    model_wrapper_actor = ray.remote(
        num_cpus=1,
        num_gpus=1,
    )(
        AttentionWrapper,
    ).options(runtime_env={"env_vars": {"KINETO_LOG_LEVEL": "5"}})

    model_wrappers = [
        model_wrapper_actor.remote(
            model_config,
            parallel_config,
            max_num_blocks,
            args.max_seq_len,
            args.block_size,
            args.attention_backend,
            dtype,
        )
        for _ in range(args.num_gpus)
    ]

    for attention_input in input_combinations:
        worker_id = len(promises)
        promise = model_wrappers[worker_id].profile.remote(attention_input)
        promises.append(promise)

        if len(promises) >= args.num_gpus:
            results = ray.get(promises)
            all_results.extend(results)
            promises = []

        pbar.update(1)

    results = ray.get(promises)
    all_results.extend(results)

    # filter all none results
    all_results = list(filter(None, all_results))

    df = pd.DataFrame(all_results)
    # the time_stats column is a dict, so we need to expand it into columns recursively and add prefix
    df = (
        pd.json_normalize(df["time_stats"])
        .add_prefix("time_stats.")
        .join(df.drop(columns=["time_stats"]))
    )
    return df


def main():
    args = parse_args()

    dtype = torch.float16
    input_combinations = get_attention_input_combinations(
        args.max_seq_len,
        args.min_batch_size,
        args.max_batch_size,
        args.profile_only_prefill,
        args.profile_only_decode,
        args.max_chunk_size,
    )

    total_combos = {}
    max_num_blocks_dict = {}
    for model in args.models:
        model_config = ModelConfig.from_model_name(
            model, max_model_len=args.max_seq_len
        )
        for num_tensor_parallel_workers in args.num_tensor_parallel_workers:
            max_num_blocks = get_max_num_blocks(
                model_config,
                ParallelConfig(
                    tensor_parallel_size=num_tensor_parallel_workers,
                    pipeline_parallel_size=1,
                ),
                args.block_size,
                dtype,
            )
            max_num_blocks_dict[(model, num_tensor_parallel_workers)] = max_num_blocks
            total_combos[(model, num_tensor_parallel_workers)] = list(
                filter(
                    lambda input_combination: input_combination.is_under_memory_limit(
                        max_num_blocks * args.block_size
                    ),
                    input_combinations,
                )
            )

    pbar = tqdm(total=sum(len(v) for v in total_combos.values()))

    for model in args.models:
        result_df = pd.DataFrame()
        for num_tensor_parallel_workers in args.num_tensor_parallel_workers:
            result_df = pd.concat(
                [
                    result_df,
                    profile_model(
                        args,
                        model,
                        num_tensor_parallel_workers,
                        total_combos[(model, num_tensor_parallel_workers)],
                        max_num_blocks_dict[(model, num_tensor_parallel_workers)],
                        dtype,
                        pbar,
                    ),
                ]
            )
        # model name would contain '/', so create a directory as required
        os.makedirs(f"{args.output_dir}/{model}", exist_ok=True)
        result_df.to_csv(
            f"{args.output_dir}/{model}/attention.csv", index=False, compression="xz"
        )


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import Any

from vidur.types import BaseIntEnum


class BaseRegistry(ABC):
    _key_class = BaseIntEnum

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._registry = {}

    @classmethod
    def register(cls, key: BaseIntEnum, implementation_class: Any) -> None:
        if key in cls._registry:
            return

        cls._registry[key] = implementation_class

    @classmethod
    def unregister(cls, key: BaseIntEnum) -> None:
        if key not in cls._registry:
            raise ValueError(f"{key} is not registered")

        del cls._registry[key]

    @classmethod
    def get(cls, key: BaseIntEnum, *args, **kwargs) -> Any:
        if key not in cls._registry:
            raise ValueError(f"{key} is not registered")

        return cls._registry[key](*args, **kwargs)

    @classmethod
    def get_class(cls, key: BaseIntEnum) -> Any:
        if key not in cls._registry:
            raise ValueError(f"{key} is not registered")

        return cls._registry[key]

    @classmethod
    @abstractmethod
    def get_key_from_str(cls, key_str: str) -> BaseIntEnum:
        pass

    @classmethod
    def get_from_str(cls, key_str: str, *args, **kwargs) -> Any:
        return cls.get(cls.get_key_from_str(key_str), *args, **kwargs)



from typing import List, Tuple

from vidur.entities import Request
from vidur.scheduler.global_scheduler.base_global_scheduler import BaseGlobalScheduler


class LORGlobalScheduler(BaseGlobalScheduler):
    """
    Least outstanding requests (LOR) global scheduler.
    """

    def schedule(self) -> List[Tuple[int, Request]]:
        self.sort_requests()

        request_mapping = []
        # keep a map of replica_id -> replica_scheduler
        # this is used to find the replica with the least outstanding requests
        pending_requests_map = {
            replica_scheduler.replica_id: replica_scheduler.num_pending_requests
            for replica_scheduler in self._replica_schedulers.values()
        }

        # using a very simple implementation here, to keep wiring simple
        while self._request_queue:
            request = self._request_queue.pop(0)
            replica_id = min(pending_requests_map.items(), key=lambda x: x[1])[0]
            pending_requests_map[replica_id] += 1
            request_mapping.append((replica_id, request))

        return request_mapping



import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import streamlit as st
from paretoset import paretoset

from vidur.config_optimizer.analyzer.bottleneck_analyzer import BottleneckAnalyzer
from vidur.config_optimizer.analyzer.constants import AXIS_COLS
from vidur.config_optimizer.analyzer.dashboard.utils import (
    add_advanced_filters,
    add_model_trace_selector,
    add_small_divider,
    write_best_config,
)


def plot_pareto_curve(df, metric, percentile, slo):
    metric_col = f"{metric}_{percentile}%"

    # filter out points which are more than 1.5x away from the SLO limit
    df = df[(df[metric_col] <= slo * 1.5)]

    paretoset_mask = paretoset(df[[metric_col, "QPS per Dollar"]], sense=["min", "max"])
    pareto_df = df[paretoset_mask]

    fig = go.Figure()

    # Scatter plot for all configurations
    configs_trace = px.scatter(
        df,
        x=metric_col,
        y="QPS per Dollar",
        hover_data=list(AXIS_COLS.values()),
    )
    # set the color of the scatter plot to blue
    configs_trace["data"][0]["marker"]["color"] = "blue"
    fig.add_trace(configs_trace["data"][0])

    # Pareto frontier
    # sort the pareto_df by metric_col
    pareto_df = pareto_df.sort_values(metric_col)

    pareto_trace = px.line(
        pareto_df,
        x=metric_col,
        y="QPS per Dollar",
        hover_data=list(AXIS_COLS.values()),
    )
    pareto_trace["data"][0]["line"]["color"] = "orange"
    fig.add_trace(pareto_trace["data"][0])

    # SLO line
    fig.add_vline(
        x=slo,
        line=dict(
            color="Red",
        ),
        name="SLO Limit",
    )

    # add a vrect for the SLO compliant region
    fig.add_vrect(
        x0=0, x1=slo, fillcolor="green", opacity=0.05, layer="below", line_width=0
    )

    # set x limit to slo * 1.25 times
    fig.update_xaxes(range=[0, slo * 1.25])

    # Layout
    fig.update_layout(
        title=f"Pareto Curve for {metric} vs. QPS per Dollar",
        xaxis_title=f"{metric} - P{percentile}",
        yaxis_title="QPS per Dollar",
        legend=dict(
            yanchor="top",
            y=0.99,
            xanchor="left",
            x=0.01,
        ),
    )

    # Display
    st.plotly_chart(fig, use_container_width=True)


def plot_metrics_scatter(
    df, metric_1, percentile_1, slo_1, metric_2, percentile_2, slo_2
):
    metric_1_col = f"{metric_1}_{percentile_1}%"
    metric_2_col = f"{metric_2}_{percentile_2}%"

    slo_compliant_df = df[(df[metric_1_col] <= slo_1) & (df[metric_2_col] <= slo_2)]
    best_config = slo_compliant_df.sort_values("QPS per Dollar", ascending=False).iloc[
        0
    ]
    # convert the best config to a dataframe
    best_config = pd.DataFrame([best_config])

    fig = go.Figure()

    configs_trace = px.scatter(
        df,
        x=metric_1_col,
        y=metric_2_col,
        hover_data=list(AXIS_COLS.values()),
        color="QPS per Dollar",
    )
    fig.add_trace(configs_trace["data"][0])

    # Best configuration
    best_config_trace = px.scatter(
        best_config,
        x=metric_1_col,
        y=metric_2_col,
        hover_data=list(AXIS_COLS.values()),
    )
    # set the color of the best config to orange
    best_config_trace["data"][0]["marker"]["color"] = "orange"
    # set maker to star
    best_config_trace["data"][0]["marker"]["symbol"] = "star"
    # enlarge the size of the marker
    best_config_trace["data"][0]["marker"]["size"] = 12
    fig.add_trace(best_config_trace["data"][0])

    # SLO lines
    fig.add_shape(
        type="line",
        x0=slo_1,
        y0=0,
        y1=slo_2 * 1.25,
        x1=slo_1,
        line=dict(color="Red"),
        name=f"SLO Limit for {metric_1}",
    )
    fig.add_shape(
        type="line",
        x0=0,
        x1=slo_1 * 1.25,
        y0=slo_2,
        y1=slo_2,
        line=dict(color="Red"),
        name=f"SLO Limit for {metric_2}",
    )

    # set x and y limits
    fig.update_xaxes(range=[0, slo_1 * 1.25])
    fig.update_yaxes(range=[0, slo_2 * 1.25])

    # Highlighting SLO compliant area
    fig.add_shape(
        type="rect",
        x0=0,
        y0=0,
        x1=slo_1,
        y1=slo_2,
        fillcolor="green",
        opacity=0.05,
        layer="below",
        line_width=0,
    )

    # Layout
    fig.update_layout(
        title=f"{metric_1} vs. {metric_2} Colored by QPS per Dollar",
        xaxis_title=f"{metric_1} - P{percentile_1}",
        yaxis_title=f"{metric_2} - P{percentile_2}",
        coloraxis_colorbar=dict(
            title="QPS per Dollar",
            title_side="right",
        ),
    )

    # Display
    st.plotly_chart(fig, use_container_width=True)


def plot_pareto_curve_under_slos(
    bottleneck_analyzer,
    df,
    metric_1,
    percentile_1,
    slo_1,
    metric_2,
    percentile_2,
    slo_2,
):
    write_best_config(
        bottleneck_analyzer,
        df,
        metric_1,
        percentile_1,
        slo_1,
        metric_2,
        percentile_2,
        slo_2,
    )

    add_small_divider()

    col1, col2, col3 = st.columns(3)
    with col1:
        plot_pareto_curve(df, metric_1, percentile_1, slo_1)
    with col2:
        plot_pareto_curve(df, metric_2, percentile_2, slo_2)
    with col3:
        plot_metrics_scatter(
            df, metric_1, percentile_1, slo_1, metric_2, percentile_2, slo_2
        )


def render_pareto_curve_page(subset_dfs):
    # create a subpage for best config selection
    st.markdown("## Pareto Curve Analysis")
    st.markdown(
        "This tool helps you visualize the trade-offs between different performance metrics and helps you identify the best configurations."
    )
    add_small_divider()
    st.markdown("### Input Parameters")

    model_select_box, trace_select_box = add_model_trace_selector(subset_dfs)

    subset_df = subset_dfs[(model_select_box, trace_select_box)]

    col1, col2 = st.columns(2)

    with col1:
        ttft_slo_percentile = st.selectbox(
            "TTFT SLO Percentile:", [50, 75, 90, 95, 99], index=2
        )
        ttft_slo_value = st.slider(
            "TTFT SLO Value (s)", min_value=0.1, max_value=10.0, value=2.0
        )

    with col2:
        tbt_slo_percentile = st.selectbox(
            "TBT SLO Percentile:", [50, 75, 90, 95, 99], index=4
        )
        tbt_slo_value = st.slider(
            "TBT SLO Value (ms)", min_value=10, max_value=2000, value=200
        )

    filtered_df = add_advanced_filters(subset_df)

    bottleneck_analyzer = BottleneckAnalyzer(
        ttft_slo_percentile,
        ttft_slo_value,
        tbt_slo_percentile,
        tbt_slo_value,
    )

    st.markdown("### Results")
    add_small_divider()

    plot_pareto_curve_under_slos(
        bottleneck_analyzer,
        filtered_df,
        "TTFT",
        ttft_slo_percentile,
        ttft_slo_value,
        "TBT",
        tbt_slo_percentile,
        tbt_slo_value,
    )



from typing import Tuple

from vidur.request_generator.base_request_length_generator import (
    BaseRequestLengthGenerator,
)


class FixedRequestLengthGenerator(BaseRequestLengthGenerator):

    def get_next_num_tokens(self) -> Tuple[float, float]:
        return (
            self.config.prefill_tokens,
            self.config.decode_tokens,
        )



from collections import defaultdict
from typing import Optional

import numpy as np
import pandas as pd
import plotly_express as px
import wandb

from vidur.logger import init_logger

logger = init_logger(__name__)


class DataSeries:
    def __init__(
        self,
        x_name: str,
        y_name: str,
        subsamples: Optional[int] = None,
        save_table_to_wandb: bool = True,
        save_plots: bool = True,
    ) -> None:
        # metrics are a data series of two-dimensional (x, y) datapoints
        self._data_series = []
        # column names of x, y datatpoints for data collection
        self._x_name = x_name
        self._y_name = y_name

        # most recently collected y datapoint for incremental updates
        # to aid incremental updates to y datapoints
        self._last_data_y = 0

        self._subsamples = subsamples
        self._save_table_to_wandb = save_table_to_wandb
        self._save_plots = save_plots

    def consolidate(
        self,
    ):
        res = defaultdict(list)
        for x, y in self._data_series:
            res[x].append(y)
        self._data_series = [(x, sum(y) / len(y)) for x, y in res.items()]

        # sort by x
        self._data_series = sorted(self._data_series, key=lambda x: x[0])
        self._last_data_y = self._data_series[-1][1] if len(self._data_series) else 0

    def __len__(self):
        return len(self._data_series)

    @property
    def _metric_name(self) -> str:
        return self._y_name

    # add a new x, y datapoint
    def put(self, data_x: float, data_y: float) -> None:
        self._last_data_y = data_y
        self._data_series.append((data_x, data_y))

    # get most recently collected y datapoint
    def _peek_y(self):
        return self._last_data_y

    # convert list of x, y datapoints to a pandas dataframe
    def _to_df(self):
        return pd.DataFrame(self._data_series, columns=[self._x_name, self._y_name])

    # add a new x, y datapoint as an incremental (delta) update to
    # recently collected y datapoint
    def put_delta(self, data_x: float, data_y_delta: float) -> None:
        last_data_y = self._peek_y()
        data_y = last_data_y + data_y_delta
        self.put(data_x, data_y)

    def print_series_stats(
        self, df: pd.DataFrame, plot_name: str, x_name: str = None, y_name: str = None
    ) -> None:
        if len(self._data_series) == 0:
            return
        if x_name is None:
            x_name = self._x_name

        if y_name is None:
            y_name = self._y_name

        logger.debug(
            f"{plot_name}: {y_name} stats:"
            f" min: {df[y_name].min()},"
            f" max: {df[y_name].max()},"
            f" mean: {df[y_name].mean()},"
        )
        if wandb.run:
            wandb.log(
                {
                    f"{plot_name}_min": df[y_name].min(),
                    f"{plot_name}_max": df[y_name].max(),
                    f"{plot_name}_mean": df[y_name].mean(),
                },
                step=0,
            )

    def print_distribution_stats(
        self, df: pd.DataFrame, plot_name: str, y_name: str = None
    ) -> None:
        if len(self._data_series) == 0:
            return

        if y_name is None:
            y_name = self._y_name

        logger.debug(
            f"{plot_name}: {y_name} stats:"
            f" min: {df[y_name].min()},"
            f" max: {df[y_name].max()},"
            f" mean: {df[y_name].mean()},"
            f" median: {df[y_name].median()},"
            f" 95th percentile: {df[y_name].quantile(0.95)},"
            f" 99th percentile: {df[y_name].quantile(0.99)}"
            f" 99.9th percentile: {df[y_name].quantile(0.999)}"
        )
        if wandb.run:
            wandb.log(
                {
                    f"{plot_name}_min": df[y_name].min(),
                    f"{plot_name}_max": df[y_name].max(),
                    f"{plot_name}_mean": df[y_name].mean(),
                    f"{plot_name}_median": df[y_name].median(),
                    f"{plot_name}_95th_percentile": df[y_name].quantile(0.95),
                    f"{plot_name}_99th_percentile": df[y_name].quantile(0.99),
                    f"{plot_name}_99.9th_percentile": df[y_name].quantile(0.999),
                },
                step=0,
            )

    def _should_subsample(self, length: int) -> bool:
        return isinstance(self._subsamples, int) and length > self._subsamples

    def _save_df(self, df: pd.DataFrame, path: str, plot_name: str) -> None:
        df.to_csv(f"{path}/{plot_name}.csv")
        if wandb.run and self._save_table_to_wandb:
            wand_table = wandb.Table(dataframe=df)
            wandb.log({f"{plot_name}_table": wand_table}, step=0)

    def save_df(self, path: str, plot_name: str) -> None:
        df = self._to_df()
        self._save_df(df, path, plot_name)

    def plot_step(
        self,
        path: str,
        plot_name: str,
        y_axis_label: str = None,
        start_time: float = 0,
        y_cumsum: bool = True,
    ) -> None:

        if len(self._data_series) == 0:
            return

        if y_axis_label is None:
            y_axis_label = self._y_name

        df = self._to_df()
        df[self._x_name] -= start_time

        if y_cumsum:
            df[self._y_name] = df[self._y_name].cumsum()

        self.print_series_stats(df, plot_name)

        # subsample
        if self._should_subsample(len(df)):
            # pick self._subsamples from the dataframe
            # however, if we make the difference between indices constant
            # we might pick spurious periodic patterns
            # so we pick the indices with a constant difference in the indices with a random offset
            # this is to avoid picking spurious periodic patterns
            indices = np.arange(0, len(df), len(df) // self._subsamples)
            offsets = np.random.randint(0, 5)
            indices = (indices + offsets) % len(df)
            df = df.iloc[indices]

        if wandb.run:
            wandb_df = df.copy()
            # rename the self._y_name column to y_axis_label
            wandb_df = wandb_df.rename(columns={self._y_name: y_axis_label})

            wandb.log(
                {
                    f"{plot_name}_step": wandb.plot.line(
                        wandb.Table(dataframe=wandb_df),
                        self._x_name,
                        y_axis_label,
                        title=plot_name,
                    )
                },
                step=0,
            )

        if self._save_plots:
            fig = px.line(
                df,
                x=self._x_name,
                y=self._y_name,
                markers=True,
                labels={"x": y_axis_label},
            )
            fig.update_traces(marker=dict(color="red", size=2))
            fig.write_image(f"{path}/{plot_name}.png")

        self._save_df(df, path, plot_name)

    def plot_cdf(self, path: str, plot_name: str, y_axis_label: str = None) -> None:
        if len(self._data_series) == 0:
            return

        if y_axis_label is None:
            y_axis_label = self._y_name

        df = self._to_df()

        self.print_distribution_stats(df, plot_name)

        df["cdf"] = df[self._y_name].rank(method="first", pct=True)
        # sort by cdf
        df = df.sort_values(by=["cdf"])

        # subsample
        if self._should_subsample(len(df)):
            df = df.iloc[:: len(df) // self._subsamples]

        if wandb.run:
            wandb_df = df.copy()
            # rename the self._y_name column to y_axis_label
            wandb_df = wandb_df.rename(columns={self._y_name: y_axis_label})

            wandb.log(
                {
                    f"{plot_name}_cdf": wandb.plot.line(
                        wandb.Table(dataframe=wandb_df),
                        "cdf",
                        y_axis_label,
                        title=plot_name,
                    )
                },
                step=0,
            )

        if self._save_plots:
            fig = px.line(
                df, x=self._y_name, y="cdf", markers=True, labels={"x": y_axis_label}
            )
            fig.update_traces(marker=dict(color="red", size=2))
            fig.write_image(f"{path}/{plot_name}.png")
        self._save_df(df, path, plot_name)

    def plot_histogram(self, path: str, plot_name: str) -> None:
        if len(self._data_series) == 0:
            return

        df = self._to_df()

        self.print_distribution_stats(df, plot_name)

        if wandb.run:
            # wandb histogram is highly inaccurate so we need to generate the histogram
            # ourselves and then use wandb bar chart
            histogram_df = (
                df[self._y_name].value_counts(bins=25, sort=False).sort_index()
            )
            histogram_df = histogram_df.reset_index()
            histogram_df.columns = ["Bins", "count"]
            histogram_df["Bins"] = histogram_df["Bins"].apply(lambda x: x.mid)
            histogram_df = histogram_df.sort_values(by=["Bins"])
            # convert to percentage
            histogram_df["Percentage"] = histogram_df["count"] * 100 / len(df)
            # drop bins with less than 0.1% of the total count
            histogram_df = histogram_df[histogram_df["Percentage"] > 0.1]

            wandb.log(
                {
                    f"{plot_name}_histogram": wandb.plot.bar(
                        wandb.Table(dataframe=histogram_df),
                        "Bins",
                        "Percentage",  # wandb plots are horizontal
                        title=plot_name,
                    )
                },
                step=0,
            )

        if self._save_plots:
            fig = px.histogram(df, x=self._y_name, nbins=25)
            fig.write_image(f"{path}/{plot_name}.png")

    def plot_differential(self, path: str, plot_name: str) -> None:
        if len(self._data_series) == 0:
            return

        df = self._to_df()

        differential_col_name = f"{self._y_name}/{self._x_name}"

        # if there are repeated values of x, then we cannot compute a differential
        # in that, case aggreagate the y values for the same x value
        df = df.groupby(self._x_name).max().reset_index().sort_values(self._x_name)

        # compute differential
        df[differential_col_name] = df[self._y_name].diff() / df[self._x_name].diff()

        df = df.dropna()

        self.print_distribution_stats(df, plot_name, differential_col_name)

        # subsample
        if self._should_subsample(len(df)):
            df = df.iloc[:: len(df) // self._subsamples]

        if wandb.run:
            wandb.log(
                {
                    f"{plot_name}_differential": wandb.plot.line(
                        wandb.Table(dataframe=df),
                        self._x_name,
                        differential_col_name,
                        title=plot_name,
                    )
                },
                step=0,
            )

        if self._save_plots:
            fig = px.line(df, x=self._x_name, y=differential_col_name, markers=True)
            fig.update_traces(marker=dict(color="red", size=2))
            fig.write_image(f"{path}/{plot_name}.png")

        self._save_df(df, path, plot_name)



from paretoset import paretoset


def get_near_pareto_mask(df, metric_col, percentile, tolerance=0.1):
    metric_col = f"{metric_col}_{percentile}"

    paretoset_mask = paretoset(
        df[["capacity_per_dollar", metric_col]], sense=["max", "min"]
    )
    pareto_df = df[paretoset_mask]

    # iterate over all the rows in pareto_df and find the rows that are within tolerance of the pareto frontier
    near_pareto_mask = df["cost"] == -1

    for _, row in pareto_df.iterrows():
        # find all the rows in df that are within tolerance of the pareto frontier
        near_pareto_mask = near_pareto_mask | (
            (df["capacity_per_dollar"] >= row["capacity_per_dollar"] * (1 - tolerance))
            & (df[metric_col] <= row[metric_col] * (1 + tolerance))
        )

    return near_pareto_mask


def get_trace_name(trace_file: str) -> str:
    if "arxiv_summarization" in trace_file:
        return "ArxivSum"
    elif "lmsys_chat_1m" in trace_file:
        return "Chat1M"
    elif "bwb" in trace_file:
        return "BWB"
    else:
        return trace_file.split("/")[-1].split(".")[0]



import json
import uuid

import numpy as np
import torch


class RecordFunctionTracer:
    def __init__(self, output_path: str):
        trace_id = str(uuid.uuid4())[:8]
        self.trace_path = (
            f"{output_path}/profiler_traces/profiler_trace_{trace_id}.json"
        )

    def __enter__(self):
        self.profiler = torch.profiler.profile(
            activities=[
                torch.profiler.ProfilerActivity.CPU,
                torch.profiler.ProfilerActivity.CUDA,
            ],
        )
        self.profiler.__enter__()

    def __exit__(self, *args):
        self.profiler.__exit__(None, None, None)
        torch.cuda.synchronize()
        self.profiler.export_chrome_trace(self.trace_path)

    def find_children(self, trace, event):
        if not ("dur" in event and "ts" in event):
            return

        children = []
        for e in trace:
            if not ("dur" in e and "ts" in e):
                continue

            # if the ts of the child is completely within the ts of the parent
            if (
                e["ts"] > event["ts"]
                and e["ts"] + e["dur"] < event["ts"] + event["dur"]
            ):
                children.append(e)
        return children

    def find_correlated_event(self, trace, event):
        if not ("args" in event and "correlation" in event["args"]):
            return

        for e in trace:
            if not ("args" in e and "correlation" in e["args"]):
                continue

            if e == event:
                continue

            if e["args"]["correlation"] == event["args"]["correlation"]:
                return e

    def get_operation_time_stats(self):
        stats = {}

        trace = json.load(open(self.trace_path, "r"))["traceEvents"]

        for event in trace:
            if not ("cat" in event and event["cat"] == "user_annotation"):
                continue
            children = self.find_children(trace, event)
            cuda_time = 0
            for child in children:
                if not ("cat" in child and child["cat"] == "cuda_runtime"):
                    continue
                correlated_event = self.find_correlated_event(trace, child)
                if not correlated_event:
                    continue
                cuda_time += correlated_event["dur"]
            if cuda_time == 0:
                continue

            name = event["name"].replace("vidur_", "")

            if name not in stats:
                stats[name] = []

            stats[name].append(cuda_time * 1e-3)  # to convert to ms

        return {
            operation: {
                "min": np.min(times),
                "max": np.max(times),
                "mean": np.mean(times),
                "median": np.median(times),
                "std": np.std(times),
            }
            for operation, times in stats.items()
        }



from vidur.config import SimulationConfig
from vidur.simulator import Simulator
from vidur.utils.random import set_seeds


def main() -> None:
    config: SimulationConfig = SimulationConfig.create_from_cli_args()
    set_seeds(config.seed)

    simulator = Simulator(config)
    simulator.run()


if __name__ == "__main__":
    main()



import re


def sanitize_name(name: str) -> str:
    """Convert to lowercase and replace special characters with hyphens"""
    return re.sub(r"[^a-zA-Z0-9]", "-", name).lower()


def parse_metadata(metadata: list) -> dict:
    """Parse metadata key-value pairs of the form key=value"""
    metadata_dict = {}
    for item in metadata:
        if "=" not in item:
            raise ValueError(f"Metadata must be in key=value format. Got: {item}")
        key, value = item.split("=", 1)
        metadata_dict[key.strip()] = value.strip()
    return metadata_dict



from dataclasses import dataclass
from typing import Any, Dict, Optional

from vidur.config.base_fixed_config import BaseFixedConfig
from vidur.logger import init_logger
from vidur.types import ActivationType, NormType

logger = init_logger(__name__)


@dataclass
class BaseModelConfig(BaseFixedConfig):
    num_layers: int
    num_q_heads: int
    num_kv_heads: int
    embedding_dim: int
    mlp_hidden_dim: int
    max_model_len: int
    use_gated_mlp: bool
    use_bias: bool
    use_qkv_bias: bool
    activation: ActivationType
    norm: NormType
    post_attn_norm: bool
    vocab_size: int
    is_neox_style: Optional[bool] = True
    rope_theta: Optional[float] = None
    rope_scaling: Optional[Dict[str, Any]] = None
    partial_rotary_factor: float = 1.0
    no_tensor_parallel: bool = False


@dataclass
class Llama2ModelConfig(BaseModelConfig):
    max_model_len: int = 16384
    use_gated_mlp: bool = True
    use_bias: bool = False
    use_qkv_bias: bool = False
    activation: ActivationType = ActivationType.SILU
    norm: NormType = NormType.RMS_NORM
    post_attn_norm: bool = True
    vocab_size: int = 32768
    is_neox_style: Optional[bool] = True
    rope_theta: Optional[float] = 10000
    rope_scaling: Optional[Dict[str, Any]] = None
    partial_rotary_factor: float = 1.0
    no_tensor_parallel: bool = False

    @staticmethod
    def get_name():
        return "meta-llama/Llama-2-Config"


@dataclass
class CodeLlama34BModelConfig(Llama2ModelConfig):
    num_layers: int = 48
    num_q_heads: int = 64
    num_kv_heads: int = 8
    embedding_dim: int = 8192
    mlp_hidden_dim: int = 22016
    rope_theta: Optional[float] = 1000000

    @staticmethod
    def get_name():
        return "codellama/CodeLlama-34b-Instruct-hf"


@dataclass
class Llama2_7BModelConfig(Llama2ModelConfig):
    num_layers: int = 32
    num_q_heads: int = 32
    num_kv_heads: int = 32
    embedding_dim: int = 4096
    mlp_hidden_dim: int = 11008
    max_model_len: int = 4096

    @staticmethod
    def get_name():
        return "meta-llama/Llama-2-7b-hf"


@dataclass
class Llama2_70BModelConfig(Llama2ModelConfig):
    num_layers: int = 80
    num_q_heads: int = 64
    num_kv_heads: int = 8
    embedding_dim: int = 8192
    mlp_hidden_dim: int = 28672
    max_model_len: int = 4096

    @staticmethod
    def get_name():
        return "meta-llama/Llama-2-70b-hf"


@dataclass
class Llama38BModelConfig(Llama2ModelConfig):
    num_layers: int = 32
    num_q_heads: int = 32
    num_kv_heads: int = 8
    embedding_dim: int = 4096
    mlp_hidden_dim: int = 14336
    max_model_len: int = 4096
    rope_theta: Optional[float] = 500000
    vocab_size: int = 128256

    @staticmethod
    def get_name():
        return "meta-llama/Llama-3-8B"


@dataclass
class Llama370BModelConfig(Llama2ModelConfig):
    num_layers: int = 80
    num_q_heads: int = 64
    num_kv_heads: int = 8
    embedding_dim: int = 8192
    mlp_hidden_dim: int = 28672
    max_model_len: int = 8192
    rope_theta: Optional[float] = 500000
    vocab_size: int = 128256

    @staticmethod
    def get_name():
        return "meta-llama/Llama-3-70B"


@dataclass
class InternLMModelConfig(Llama2ModelConfig):
    max_model_len: int = 4096
    vocab_size: int = 103168


@dataclass
class InternLM_20BModelConfig(InternLMModelConfig):
    num_layers: int = 60
    num_q_heads: int = 40
    num_kv_heads: int = 40
    embedding_dim: int = 5120
    mlp_hidden_dim: int = 13824

    @staticmethod
    def get_name():
        return "internlm/internlm-20b"


@dataclass
class InternLM2ModelConfig(Llama2ModelConfig):
    max_model_len: int = 32768
    vocab_size: int = 92544


@dataclass
class InternLM2_20BModelConfig(InternLM2ModelConfig):
    num_layers: int = 48
    num_q_heads: int = 48
    num_kv_heads: int = 8
    embedding_dim: int = 6144
    mlp_hidden_dim: int = 16384
    rope_theta: Optional[float] = 1000000

    @staticmethod
    def get_name():
        return "internlm/internlm2-20b"


@dataclass
class Phi2ModelConfig(Llama2ModelConfig):
    num_layers: int = 32
    num_q_heads: int = 32
    num_kv_heads: int = 32
    embedding_dim: int = 2560
    mlp_hidden_dim: int = 10240
    max_model_len: int = 2048
    use_gated_mlp: bool = False
    use_bias: bool = True
    use_qkv_bias: bool = True
    activation: ActivationType = ActivationType.GELU
    norm: NormType = NormType.LAYER_NORM
    post_attn_norm: bool = False
    vocab_size: int = 51200
    rope_scaling: Optional[Dict[str, Any]] = None
    rope_theta: Optional[float] = 10000
    partial_rotary_factor: float = 0.4
    no_tensor_parallel: bool = True

    @staticmethod
    def get_name():
        return "microsoft/phi-2"


@dataclass
class QwenModelConfig(Llama2ModelConfig):
    use_qkv_bias: bool = True
    max_model_len: int = 32768
    vocab_size: int = 152064

    @staticmethod
    def get_name():
        return "Qwen/Qwen-Config"


@dataclass
class Qwen72BModelConfig(QwenModelConfig):
    num_layers: int = 80
    num_q_heads: int = 64
    num_kv_heads: int = 64
    embedding_dim: int = 8192
    mlp_hidden_dim: int = 24576
    rope_theta: Optional[float] = 1000000

    @staticmethod
    def get_name():
        return "Qwen/Qwen-72B"



from vidur.request_generator.fixed_request_length_generator import (
    FixedRequestLengthGenerator,
)
from vidur.request_generator.trace_request_length_generator import (
    TraceRequestLengthGenerator,
)
from vidur.request_generator.uniform_request_length_generator import (
    UniformRequestLengthGenerator,
)
from vidur.request_generator.zipf_request_length_generator import (
    ZipfRequestLengthGenerator,
)
from vidur.types import RequestLengthGeneratorType
from vidur.utils.base_registry import BaseRegistry


class RequestLengthGeneratorRegistry(BaseRegistry):
    pass


RequestLengthGeneratorRegistry.register(
    RequestLengthGeneratorType.ZIPF, ZipfRequestLengthGenerator
)
RequestLengthGeneratorRegistry.register(
    RequestLengthGeneratorType.UNIFORM, UniformRequestLengthGenerator
)
RequestLengthGeneratorRegistry.register(
    RequestLengthGeneratorType.TRACE, TraceRequestLengthGenerator
)
RequestLengthGeneratorRegistry.register(
    RequestLengthGeneratorType.FIXED, FixedRequestLengthGenerator
)



from vidur.config_optimizer.config_explorer.config.config import (
    ClusterConfig,
    JobConfig,
    ModelConfig,
    SchedulerConfig,
    SimulationConfig,
    TraceConfig,
)

__all__ = [
    JobConfig,
    SimulationConfig,
    TraceConfig,
    SchedulerConfig,
    ClusterConfig,
    ModelConfig,
]



import logging
from typing import Tuple

import numpy as np
import pandas as pd

from vidur.config import TraceRequestLengthGeneratorConfig
from vidur.request_generator.base_request_length_generator import (
    BaseRequestLengthGenerator,
)

logger = logging.getLogger(__name__)


class TraceRequestLengthGenerator(BaseRequestLengthGenerator):

    def __init__(self, config: TraceRequestLengthGeneratorConfig):
        super().__init__(config)

        self.trace_df = pd.read_csv(config.trace_file)

        # scale prefill and decode tokens
        self.trace_df["num_prefill_tokens"] = (
            self.trace_df["num_prefill_tokens"] * config.prefill_scale_factor
        )
        self.trace_df["num_decode_tokens"] = (
            self.trace_df["num_decode_tokens"] * config.decode_scale_factor
        )

        # make sure all the prefill and decode counts are integers
        self.trace_df["num_prefill_tokens"] = self.trace_df[
            "num_prefill_tokens"
        ].astype(int)
        self.trace_df["num_decode_tokens"] = self.trace_df["num_decode_tokens"].astype(
            int
        )

        # make sure the total does not exceed the max tokens, adjust the prefill tokens if needed
        total_tokens = (
            self.trace_df["num_prefill_tokens"] + self.trace_df["num_decode_tokens"]
        )
        diff_tokens = total_tokens - config.max_tokens
        diff_tokens = diff_tokens.clip(lower=0)

        # deduct the diff tokens from the prefill and decode tokens proportionally
        prefill_tokens_ratio = self.trace_df["num_prefill_tokens"] / total_tokens
        decode_tokens_ratio = self.trace_df["num_decode_tokens"] / total_tokens

        self.trace_df["num_prefill_tokens"] -= (
            np.ceil(diff_tokens * prefill_tokens_ratio)
        ).astype(int)

        self.trace_df["num_decode_tokens"] -= (
            np.ceil(diff_tokens * decode_tokens_ratio)
        ).astype(int)

        # make sure that there is at least one prefill and decode token
        self.trace_df["num_prefill_tokens"] = self.trace_df["num_prefill_tokens"].clip(
            lower=1
        )
        self.trace_df["num_decode_tokens"] = self.trace_df["num_decode_tokens"].clip(
            lower=1
        )

        assert all(
            self.trace_df["num_prefill_tokens"] + self.trace_df["num_decode_tokens"]
            <= self.config.max_tokens
        )

        assert all(self.trace_df["num_prefill_tokens"] > 0)

        assert all(self.trace_df["num_decode_tokens"] > 0)

        # compute pd ratio and log the 25, 50, 75, 90, 95, 99 percentiles
        pd_ratio = (
            self.trace_df["num_prefill_tokens"] / self.trace_df["num_decode_tokens"]
        )
        logger.info(
            f"Loaded request length trace file {config.trace_file} with {len(self.trace_df)} requests"
        )
        pd_distribution = pd_ratio.describe(
            percentiles=[0.25, 0.5, 0.75, 0.9, 0.95, 0.99]
        )
        logger.debug(f"Prompt/decode token ratio stats\n: {pd_distribution}")

        # randomly shuffle the df based on the seed
        self.trace_df = self.trace_df.sample(frac=1, random_state=self.config.seed)
        self.next_request_idx = 0

    def get_next_num_tokens(self) -> Tuple[float, float]:
        if self.next_request_idx >= len(self.trace_df):
            return None, None

        row = self.trace_df.iloc[self.next_request_idx]
        self.next_request_idx += 1

        return (
            row["num_prefill_tokens"],
            row["num_decode_tokens"],
        )



from typing import List, Tuple

import numpy as np

from vidur.entities.batch import Batch, Request
from vidur.scheduler.replica_scheduler.base_replica_scheduler import (
    BaseReplicaScheduler,
)


class LightLLMReplicaScheduler(BaseReplicaScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._max_micro_batch_size = self._config.batch_size_cap // self._num_stages
        assert (
            self._replica_config.block_size == 1
        ), "LightLLM scheduler only supports block size of 1."
        assert (
            self._num_stages == 1
        ), "LightLLM scheduler does not support pipeline parallel."

        self._cache_len_list = []
        self._num_waiting_iters = 0

    def on_batch_end(self, batch: Batch) -> None:
        super().on_batch_end(batch)

        for request in batch.requests:
            if request.completed:
                self.free(request.id)
            else:
                self._preempted_requests.append(request)

    def _get_tuple_tokens(self, request: Request) -> Tuple[int, int]:
        if request.scheduled:
            num_processed_tokens = request.num_processed_tokens
            remaining_tokens = (
                request.num_decode_tokens - request.num_processed_decode_tokens - 1
            )
        else:
            num_processed_tokens = request.num_prefill_tokens + 1
            remaining_tokens = request.num_decode_tokens - 1 - 1

        remaining_tokens = max(0, remaining_tokens)

        return (num_processed_tokens, remaining_tokens)

    def _can_allocate_request(self, request: Request) -> bool:
        # I have no idea what this does
        self.cache_len_list.append(self._get_tuple_tokens(request))
        self.cache_len_list.sort(key=lambda x: -x[1])

        left_out_len_array = np.array([e[1] for e in self.cache_len_list])
        # assert left_out_len_array.min() >= 0
        has_run_len_array = np.array([e[0] for e in self.cache_len_list])
        cum_run_len_array = np.cumsum(has_run_len_array)
        size_array = np.arange(1, len(self.cache_len_list) + 1, 1)

        need_max_token_num = (left_out_len_array * size_array + cum_run_len_array).max()

        return need_max_token_num < self._config.num_blocks

    def _allocate_request(self, request: Request) -> None:
        if request.id not in self._allocation_map:
            self.allocate(request.id, request.num_prefill_tokens)
            return

        self.allocate(request.id, 1)

    def _get_prefill_batch(self) -> Batch:
        requests = []
        num_q_tokens = []
        num_kv_tokens = []
        total_num_q_tokens = 0

        self.cache_len_list = [
            self._get_tuple_tokens(request) for request in self._preempted_requests
        ]

        while self._request_queue:
            request = self._request_queue[0]

            next_num_q_tokens = self._get_request_next_num_q_tokens(request)
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens

            if (
                total_num_q_tokens + next_num_q_tokens
                > self._config.max_tokens_in_batch
            ):
                break

            if len(self._allocation_map) == self._config.batch_size_cap:
                break

            if len(requests) == self._max_micro_batch_size:
                break

            if not self._can_allocate_request(request):
                break

            request = self._request_queue.pop(0)

            self._allocate_request(request)
            requests.append(request)
            num_q_tokens.append(next_num_q_tokens)
            num_kv_tokens.append(next_num_kv_tokens)
            total_num_q_tokens += next_num_q_tokens

        if requests:
            return Batch(self._replica_id, requests, num_q_tokens, num_kv_tokens)

        return

    def _get_decode_batch(self) -> Batch:
        requests = []
        num_q_tokens = []
        num_kv_tokens = []

        # all preempted_requests will have prefill completed
        while self._preempted_requests:
            assert len(requests) < self._max_micro_batch_size

            request = self._preempted_requests.pop(0)

            assert self.can_allocate(1)
            self._allocate_request(request)

            next_num_q_tokens = self._get_request_next_num_q_tokens(request)
            next_num_kv_tokens = request.num_processed_tokens + next_num_q_tokens
            requests.append(request)
            num_q_tokens.append(next_num_q_tokens)
            num_kv_tokens.append(next_num_kv_tokens)

        if not requests:
            return

        return Batch(self._replica_id, requests, num_q_tokens, num_kv_tokens)

    def _can_decode(self):
        return self.can_allocate(len(self._preempted_requests))

    def _get_next_batch(self, time: float) -> Batch:
        if not self._preempted_requests:
            batch = self._get_prefill_batch()
            if batch:
                self._num_waiting_iters = 0
            return batch

        if self._num_waiting_iters >= self._config.max_waiting_iters:
            self._num_waiting_iters = 0
            batch = self._get_prefill_batch()
            if batch:
                return batch

        if self._can_decode():
            self._num_waiting_iters += 1
            return self._get_decode_batch()
        else:
            raise RuntimeError("OOM handling not implemented yet")



from math import ceil

from vidur.config import BaseRequestGeneratorConfig, ReplicaConfig
from vidur.entities.base_entity import BaseEntity
from vidur.logger import init_logger

logger = init_logger(__name__)


class Replica(BaseEntity):
    def __init__(
        self,
        replica_config: ReplicaConfig,
        generator_config: BaseRequestGeneratorConfig,
    ) -> None:
        self._id = Replica.generate_id()

        self._replica_config = replica_config
        self._model_config = replica_config.model_config
        self._device_config = replica_config.device_config
        self._generator_config = generator_config

        assert (
            self._model_config.num_layers % self._replica_config.num_pipeline_stages
            == 0
        ), f"Number of layers ({self._model_config.num_layers}) is not divisible by the number of pipeline stages ({self._replica_config.num_pipeline_stages})"
        assert (
            self._model_config.embedding_dim % self._replica_config.tensor_parallel_size
            == 0
        ), f"Embedding dimension ({self._model_config.embedding_dim}) is not divisible by the tensor parallel size ({self._replica_config.tensor_parallel_size})"

    @property
    def id(self) -> int:
        return self._id

    @property
    def num_layers(self) -> int:
        return self._model_config.num_layers

    @property
    def num_q_heads(self) -> int:
        return self._model_config.num_q_heads

    @property
    def num_kv_heads(self) -> int:
        return self._model_config.num_kv_heads

    @property
    def embedding_dim(self) -> int:
        return self._model_config.embedding_dim

    @property
    def mlp_hidden_dim(self) -> int:
        return self._model_config.mlp_hidden_dim

    @property
    def use_gated_mlp(self) -> int:
        return self._model_config.use_gated_mlp

    @property
    def vocab_size(self) -> int:
        return self._model_config.vocab_size

    @property
    def num_pipeline_stages(self) -> int:
        return self._replica_config.num_pipeline_stages

    @property
    def num_layers_per_pipeline_stage(self) -> int:
        return self._model_config.num_layers // self._replica_config.num_pipeline_stages

    @property
    def attention_head_dim(self) -> int:
        return self._model_config.embedding_dim // self._model_config.num_q_heads

    @property
    def q_heads_per_tensor_parallel_worker(self) -> int:
        return (
            self._model_config.num_q_heads // self._replica_config.tensor_parallel_size
        )

    @property
    def kv_heads_per_tensor_parallel_worker(self) -> int:
        return ceil(
            self._model_config.num_kv_heads / self._replica_config.tensor_parallel_size
        )

    @property
    def num_tensor_parallel_workers(self) -> int:
        return self._replica_config.tensor_parallel_size

    @property
    def num_kv_parallel_groups(self) -> int:
        return self._replica_config.kv_parallel_size

    @property
    def total_memory_gb(self) -> int:
        return self._device_config.total_memory_gb

    @property
    def memory_margin_fraction(self) -> float:
        return self._replica_config.memory_margin_fraction

    @property
    def max_request_tokens(self) -> int:
        return self._generator_config.max_tokens

    @property
    def per_device_flops(self) -> float:
        return self._device_config.fp16_tflops * 2**40

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "num_layers": self.num_layers,
            "num_q_heads": self.num_q_heads,
            "num_kv_heads": self.num_kv_heads,
            "embedding_dim": self.embedding_dim,
            "mlp_hidden_dim": self.mlp_hidden_dim,
            "use_gated_mlp": self.use_gated_mlp,
            "vocab_size": self.vocab_size,
            "num_pipeline_stages": self.num_pipeline_stages,
            "num_tensor_parallel_workers": self.num_tensor_parallel_workers,
        }



import os
import platform
import socket
import time
from typing import Optional

import ray


def get_ip() -> str:
    # special handling for macos
    if platform.system() == "Darwin":
        return "127.0.0.1"

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        s.connect(("10.254.254.254", 1))
        ip = s.getsockname()[0]
    except Exception:
        ip = "127.0.0.1"
    finally:
        s.close()
    return ip


def get_nodes() -> list[str]:
    cluster_resources_keys = list(ray.available_resources().keys())
    ip_addresses = [
        x
        for x in cluster_resources_keys
        if x.startswith("node:") and x != "node:__internal_head__"
    ]

    # special handling for macos, ensure that we only have one node
    if platform.system() == "Darwin":
        assert len(ip_addresses) == 1

    return ip_addresses


def run_on_each_node(func, *args, **kwargs):
    ip_addresses = get_nodes()
    remote_func = ray.remote(func)
    return ray.get(
        [
            remote_func.options(resources={ip_address: 0.1}).remote(*args, **kwargs)
            for ip_address in ip_addresses
        ]
    )


@ray.remote
class CpuAssignmentManager:
    def __init__(self):
        self._nodes = get_nodes()
        # remove "node:" prefix
        self._nodes = [node[5:] for node in self._nodes]
        self._num_cores = os.cpu_count() - 2
        self._core_mapping = {node: [False] * self._num_cores for node in self._nodes}

    def get_cpu_core_id(self) -> Optional[int]:
        for node in self._nodes:
            for i, is_core_assigned in enumerate(self._core_mapping[node]):
                if not is_core_assigned:
                    self._core_mapping[node][i] = True
                    return node, i
        return None, None

    def release_cpu_core_id(self, node: str, cpu_core_id: int) -> None:
        self._core_mapping[node][cpu_core_id] = False


class RayParallelRunner:
    def __init__(self):
        self._cpu_assignment_manager = CpuAssignmentManager.remote()

    def map(self, func, collection):
        # try to assign a core to each task
        promises = []

        remote_func = ray.remote(func)

        for item in collection:
            node = None
            cpu_core_id = None
            while node is None:
                node, cpu_core_id = ray.get(
                    self._cpu_assignment_manager.get_cpu_core_id.remote()
                )
                time.sleep(0.1)
            # launch the task
            promise = remote_func.options(resources={f"node:{node}": 0.001}).remote(
                self._cpu_assignment_manager, cpu_core_id, item
            )
            promises.append(promise)

        return ray.get(promises)



import json
import shutil
import tempfile
from datetime import datetime
from pathlib import Path

import click

from vidur.utils import parse_metadata, sanitize_name
from vidur.utils.hf_dataset_utils import upload_folder


@click.group()
def cli():
    """Vidur Data Management CLI"""
    pass


# ========================
# PUBLISH COMMANDS
# ========================


@cli.group()
def publish():
    """Publish datasets to Hugging Face Hub"""
    pass


@publish.command()
@click.option("--org", required=True, help="Target organization on HF Hub")
@click.option(
    "--collection", required=True, help="Dataset version/collection (e.g. v1.0)"
)
@click.option(
    "--model", required=True, help="Model identifier (e.g. meta-llama/Llama-2-70b-hf)"
)
@click.option("--device", required=True, help="Device SKU (e.g. h100, a100-80gb)")
@click.option(
    "--data-path",
    type=click.Path(exists=True),
    required=True,
    help="Path to model-device data directory",
)
@click.option(
    "--use-private-repo", is_flag=True, help="Make the dataset private", default=True
)
@click.option("--metadata", multiple=True, help="Additional metadata key=value pairs")
def compute(org, collection, model, device, data_path, use_private_repo, metadata):
    """Publish compute profiling data for a model-device pair"""
    # Create standardized repo identifier
    repo_id = f"{sanitize_name(org)}/{sanitize_name(collection)}-{sanitize_name(model)}-{sanitize_name(device)}"
    # Parse additional metadata
    metadata_dict = parse_metadata(metadata)

    # Create config.json
    config = {
        "schema_version": "1.0",
        "model": model,
        "device_sku": device,
        "dataset_type": "compute",
        **metadata_dict,
    }

    upload_folder(
        repo_id=repo_id,
        config=config,
        data_path=data_path,
        private=use_private_repo,
    )

    click.echo(f"✅ Published compute data to: {repo_id}")


@publish.command()
@click.option("--org", required=True, help="Target organization on HF Hub")
@click.option(
    "--collection", required=True, help="Dataset version/collection (e.g. v1.0)"
)
@click.option(
    "--sku", required=True, help="Hardware config (e.g. h100-dgx, a100-nvlink-pair)"
)
@click.option(
    "--data-path",
    type=click.Path(exists=True),
    required=True,
    help="Path to network benchmark data",
)
@click.option(
    "--use-private-repo", is_flag=True, help="Make the dataset private", default=True
)
@click.option("--metadata", multiple=True, help="Additional metadata key=value pairs")
def network(org, collection, sku, data_path, use_private_repo, metadata):
    """Publish network performance data"""
    repo_id = f"{sanitize_name(org)}/{sanitize_name(collection)}-{sanitize_name(sku)}"
    # Parse additional metadata
    metadata_dict = parse_metadata(metadata)
    # Create config.json
    config = {
        "schema_version": "1.0",
        "sku": sku,
        "dataset_type": "network",
        **metadata_dict,
    }

    upload_folder(
        repo_id=repo_id,
        config=config,
        data_path=data_path,
        private=use_private_repo,
    )

    click.echo(f"✅ Published network data to: {repo_id}")


if __name__ == "__main__":
    cli()



from vidur.types.base_int_enum import BaseIntEnum


class EventType(BaseIntEnum):
    # at any given time step, call the schedule event at the last
    # to ensure that all the requests are processed
    BATCH_STAGE_ARRIVAL = 1
    REQUEST_ARRIVAL = 2
    BATCH_STAGE_END = 3
    BATCH_END = 4
    GLOBAL_SCHEDULE = 5
    REPLICA_STAGE_SCHEDULE = 6
    REPLICA_SCHEDULE = 7



from typing import Dict

from vidur.entities.base_entity import BaseEntity


class ExecutionTime(BaseEntity):
    def __init__(
        self,
        num_layers_per_pipeline_stage: int,
        attention_rope_execution_time: float,
        attention_kv_cache_save_execution_time: float,
        attention_decode_execution_time: float,
        attention_prefill_execution_time: float,
        attention_layer_pre_proj_execution_time: float,
        attention_layer_post_proj_execution_time: float,
        mlp_layer_up_proj_execution_time: float,
        mlp_layer_down_proj_execution_time: float,
        mlp_layer_act_execution_time: float,
        attn_norm_time: float,
        mlp_norm_time: float,
        add_time: float,
        tensor_parallel_communication_time: float,
        pipeline_parallel_communication_time: float,
    ) -> None:
        self._id = ExecutionTime.generate_id()

        self._num_layers_per_pipeline_stage = num_layers_per_pipeline_stage
        self._attention_rope_execution_time = attention_rope_execution_time
        self._attention_kv_cache_save_execution_time = (
            attention_kv_cache_save_execution_time
        )
        self._attention_decode_execution_time = attention_decode_execution_time
        self._attention_prefill_execution_time = attention_prefill_execution_time
        self._attention_layer_pre_proj_execution_time = (
            attention_layer_pre_proj_execution_time
        )
        self._attention_layer_post_proj_execution_time = (
            attention_layer_post_proj_execution_time
        )
        self._mlp_layer_up_proj_execution_time = mlp_layer_up_proj_execution_time
        self._mlp_layer_down_proj_execution_time = mlp_layer_down_proj_execution_time
        self._mlp_layer_act_execution_time = mlp_layer_act_execution_time
        self._mlp_norm_time = mlp_norm_time
        self._attn_norm_time = attn_norm_time
        self._add_time = add_time
        self._tensor_parallel_communication_time = tensor_parallel_communication_time
        self._pipeline_parallel_communication_time = (
            pipeline_parallel_communication_time
        )
        # TODO(Amey): Need to add accounting for kvp group communication time
        self._kvp_group_communication_time = 0

    def _get_mlp_layer_execution_time(self) -> float:
        return (
            self._mlp_layer_up_proj_execution_time
            + self._mlp_layer_down_proj_execution_time
            + self._mlp_layer_act_execution_time
            + self._tensor_parallel_communication_time
            + self._mlp_norm_time
        )

    def _get_attention_layer_execution_time(self) -> float:
        # TODO(Amey): The KV cache save kernel in older version of
        # flashinfer had poor performance. Disabled the kv cache save
        # time for the time being. However, we need to get new profiling
        # data to re-enable this.
        return (
            self._attention_layer_pre_proj_execution_time
            + self._attention_layer_post_proj_execution_time
            + self._attention_rope_execution_time
            # + self._attention_kv_cache_save_execution_time
            + self._attention_decode_execution_time
            + self._attention_prefill_execution_time
            + self._tensor_parallel_communication_time
            + self._attn_norm_time
        )

    def _get_block_execution_time(self) -> float:
        return (
            self._get_attention_layer_execution_time()
            + self._get_mlp_layer_execution_time()
            + self._add_time
        )

    @property
    def num_layers(self) -> int:
        return self._num_layers_per_pipeline_stage

    @property
    def mlp_layer_up_proj_execution_time(self) -> float:
        return self._mlp_layer_up_proj_execution_time

    @property
    def mlp_layer_down_proj_execution_time(self) -> float:
        return self._mlp_layer_down_proj_execution_time

    @property
    def mlp_layer_act_execution_time(self) -> float:
        return self._mlp_layer_act_execution_time

    @property
    def mlp_all_reduce_time(self) -> float:
        return self._tensor_parallel_communication_time

    @property
    def attention_pre_proj_time(self) -> float:
        return self._attention_layer_pre_proj_execution_time

    @property
    def attention_post_proj_time(self) -> float:
        return self._attention_layer_post_proj_execution_time

    @property
    def attention_all_reduce_time(self) -> float:
        return self._tensor_parallel_communication_time

    @property
    def attention_rope_execution_time(self) -> float:
        return self._attention_rope_execution_time

    @property
    def attention_kv_cache_save_execution_time(self) -> float:
        return self._attention_kv_cache_save_execution_time

    @property
    def attention_decode_execution_time(self) -> float:
        return self._attention_decode_execution_time

    @property
    def attention_prefill_execution_time(self) -> float:
        return self._attention_prefill_execution_time

    @property
    def pipeline_parallel_communication_time(self) -> float:
        return self._pipeline_parallel_communication_time

    @property
    def kvp_group_communication_time(self) -> float:
        return self._kvp_group_communication_time

    @property
    def mlp_norm_time(self) -> float:
        return self._mlp_norm_time

    @property
    def attn_norm_time(self) -> float:
        return self._attn_norm_time

    @property
    def add_time(self) -> float:
        return self._add_time

    @property
    def model_time(self) -> float:
        # we are not counting the execution time for the embedding layer and last softmax layer
        block_execution_time = self._get_block_execution_time()
        pipeline_stage_execution_time = (
            block_execution_time * self._num_layers_per_pipeline_stage
        )
        # return in seconds
        return (
            pipeline_stage_execution_time
            + self.pipeline_parallel_communication_time
            + self.kvp_group_communication_time
        ) * 1e-3

    @property
    def model_time_ms(self) -> float:
        return self.model_time * 1e3

    @property
    def total_time(self) -> float:
        # return in seconds
        return self.model_time

    def to_dict(self) -> Dict[str, float]:
        return {
            "mlp_layer_up_proj_execution_time": self._mlp_layer_up_proj_execution_time,
            "mlp_layer_down_proj_execution_time": self._mlp_layer_down_proj_execution_time,
            "mlp_layer_act_execution_time": self._mlp_layer_act_execution_time,
            "mlp_norm_time": self._mlp_norm_time,
            "attention_layer_pre_proj_execution_time": self._attention_layer_pre_proj_execution_time,
            "attention_layer_post_proj_execution_time": self._attention_layer_post_proj_execution_time,
            "attention_rope_execution_time": self._attention_rope_execution_time,
            "attention_kv_cache_save_execution_time": self._attention_kv_cache_save_execution_time,
            "attention_decode_execution_time": self._attention_decode_execution_time,
            "attention_prefill_execution_time": self._attention_prefill_execution_time,
            "tensor_parallel_communication_time": self._tensor_parallel_communication_time,
            "pipeline_parallel_communication_time": self._pipeline_parallel_communication_time,
            "schedule_time": self._schedule_time,
            "attn_norm_time": self._attn_norm_time,
        }



import json
import os
from abc import ABC
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional

from vidur.config.base_poly_config import BasePolyConfig
from vidur.config.device_sku_config import BaseDeviceSKUConfig
from vidur.config.flat_dataclass import create_flat_dataclass
from vidur.config.model_config import BaseModelConfig
from vidur.config.node_sku_config import BaseNodeSKUConfig
from vidur.config.utils import dataclass_to_dict
from vidur.logger import init_logger
from vidur.types import (
    ExecutionTimePredictorType,
    GlobalSchedulerType,
    ReplicaSchedulerType,
    RequestGeneratorType,
    RequestIntervalGeneratorType,
    RequestLengthGeneratorType,
)

logger = init_logger(__name__)

ROOT_DIR = os.path.abspath(os.path.join(__file__, "../../../"))


@dataclass
class BaseRequestIntervalGeneratorConfig(BasePolyConfig):
    seed: int = field(
        default=42,
        metadata={"help": "Seed for the random number generator."},
    )


@dataclass
class BaseRequestLengthGeneratorConfig(BasePolyConfig):
    seed: int = field(
        default=42,
        metadata={"help": "Seed for the random number generator."},
    )
    max_tokens: int = field(
        default=4096,
        metadata={"help": "Maximum tokens."},
    )


@dataclass
class TraceRequestIntervalGeneratorConfig(BaseRequestIntervalGeneratorConfig):
    trace_file: str = field(
        default="data/processed_traces/AzureFunctionsInvocationTraceForTwoWeeksJan2021Processed.csv",
        metadata={"help": "Path to the trace request interval generator file."},
    )
    start_time: str = field(
        default="1970-01-04 12:00:00",
        metadata={"help": "Start time of the trace request interval generator."},
    )
    end_time: str = field(
        default="1970-01-04 15:00:00",
        metadata={"help": "End time of the trace request interval generator."},
    )
    time_scale_factor: float = field(
        default=1,
        metadata={
            "help": "Time scale factor for the trace request interval generator."
        },
    )

    @staticmethod
    def get_type():
        return RequestIntervalGeneratorType.TRACE


@dataclass
class PoissonRequestIntervalGeneratorConfig(BaseRequestIntervalGeneratorConfig):
    qps: float = field(
        default=10,
        metadata={"help": "Queries per second for Poisson Request Interval Generator."},
    )

    @staticmethod
    def get_type():
        return RequestIntervalGeneratorType.POISSON


@dataclass
class GammaRequestIntervalGeneratorConfig(BaseRequestIntervalGeneratorConfig):
    qps: float = field(
        default=0.2,
        metadata={"help": "Queries per second for Gamma Request Interval Generator."},
    )
    cv: float = field(
        default=0.5,
        metadata={
            "help": "Coefficient of variation for Gamma Request Interval Generator."
        },
    )

    @staticmethod
    def get_type():
        return RequestIntervalGeneratorType.GAMMA


@dataclass
class StaticRequestIntervalGeneratorConfig(BaseRequestIntervalGeneratorConfig):
    @staticmethod
    def get_type():
        return RequestIntervalGeneratorType.STATIC


@dataclass
class TraceRequestLengthGeneratorConfig(BaseRequestLengthGeneratorConfig):
    trace_file: str = field(
        default="data/processed_traces/sharegpt_8k_filtered_stats_llama2_tokenizer.csv",
        metadata={"help": "Path to the trace request length generator file."},
    )
    prefill_scale_factor: float = field(
        default=1,
        metadata={
            "help": "Prefill scale factor for the trace request length generator."
        },
    )
    decode_scale_factor: float = field(
        default=1,
        metadata={
            "help": "Decode scale factor for the trace request length generator."
        },
    )
    max_tokens: int = field(
        default=1024 * 1024,
        metadata={"help": "Maximum tokens for the trace request length generator."},
    )

    @staticmethod
    def get_type():
        return RequestLengthGeneratorType.TRACE


@dataclass
class ZipfRequestLengthGeneratorConfig(BaseRequestLengthGeneratorConfig):
    theta: float = field(
        default=0.6,
        metadata={"help": "Theta for Zipf Request Length Generator."},
    )
    scramble: bool = field(
        default=False,
        metadata={"help": "Scramble for Zipf Request Length Generator."},
    )
    min_tokens: int = field(
        default=1024,
        metadata={"help": "Minimum tokens for Zipf Request Length Generator."},
    )
    max_tokens: int = field(
        default=4096,
        metadata={"help": "Maximum tokens for Zipf Request Length Generator."},
    )
    prefill_to_decode_ratio: float = field(
        default=20.0,
        metadata={"help": "Prefill to decode ratio for Zipf Request Length Generator."},
    )

    @staticmethod
    def get_type():
        return RequestLengthGeneratorType.ZIPF


@dataclass
class UniformRequestLengthGeneratorConfig(BaseRequestLengthGeneratorConfig):
    min_tokens: int = field(
        default=1024,
        metadata={"help": "Minimum tokens for Uniform Request Length Generator."},
    )
    max_tokens: int = field(
        default=4096,
        metadata={"help": "Maximum tokens for Uniform Request Length Generator."},
    )
    prefill_to_decode_ratio: float = field(
        default=20.0,
        metadata={
            "help": "Prefill to decode ratio for Uniform Request Length Generator."
        },
    )

    @staticmethod
    def get_type():
        return RequestLengthGeneratorType.UNIFORM


@dataclass
class FixedRequestLengthGeneratorConfig(BaseRequestLengthGeneratorConfig):
    prefill_tokens: int = field(
        default=4 * 1024,
        metadata={"help": "Prefill tokens for Fixed Request Length Generator."},
    )
    decode_tokens: int = field(
        default=512,
        metadata={"help": "Decode tokens for Fixed Request Length Generator."},
    )

    @staticmethod
    def get_type():
        return RequestLengthGeneratorType.FIXED


@dataclass
class BaseRequestGeneratorConfig(BasePolyConfig):
    seed: int = field(
        default=42,
        metadata={"help": "Seed for the random number generator."},
    )
    max_tokens: int = field(
        default=4096,
        metadata={"help": "Maximum tokens."},
    )


@dataclass
class SyntheticRequestGeneratorConfig(BaseRequestGeneratorConfig):
    length_generator_config: BaseRequestLengthGeneratorConfig = field(
        default_factory=FixedRequestLengthGeneratorConfig,
        metadata={"help": "Length generator config for Synthetic Request Generator."},
    )
    interval_generator_config: BaseRequestIntervalGeneratorConfig = field(
        default_factory=StaticRequestIntervalGeneratorConfig,
        metadata={"help": "Interval generator config for Synthetic Request Generator."},
    )
    num_requests: Optional[int] = field(
        default=512,
        metadata={"help": "Number of requests for Synthetic Request Generator."},
    )
    duration: Optional[float] = field(
        default=None,
        metadata={"help": "Duration of the synthetic request generator."},
    )

    def __post_init__(self):
        self.max_tokens = self.length_generator_config.max_tokens

    @staticmethod
    def get_type():
        return RequestGeneratorType.SYNTHETIC


@dataclass
class TraceRequestGeneratorConfig(BaseRequestGeneratorConfig):
    trace_file: str = field(
        default="data/processed_traces/splitwise_conv.csv",
        metadata={"help": "Path to the trace request generator file."},
    )
    date: str = field(
        default="2023-08-21",
        metadata={"help": "Date for the trace request generator."},
    )
    prefill_scale_factor: float = field(
        default=0.3,
        metadata={"help": "Prefill scale factor for the trace request generator."},
    )
    decode_scale_factor: float = field(
        default=1,
        metadata={"help": "Decode scale factor for the trace request generator."},
    )
    time_scale_factor: float = field(
        default=0.04,
        metadata={"help": "Time scale factor for the trace request generator."},
    )
    max_tokens: int = field(
        default=4096,
        metadata={"help": "Maximum tokens for the trace request generator."},
    )

    @staticmethod
    def get_type():
        return RequestGeneratorType.TRACE_REPLAY


@dataclass
class BaseReplicaSchedulerConfig(BasePolyConfig):
    watermark_blocks_fraction: float = field(
        default=0.01,
        metadata={"help": "Watermark blocks fraction."},
    )
    num_blocks: Optional[int] = field(
        default=None,
        metadata={"help": "Number of blocks."},
    )
    batch_size_cap: int = field(
        default=128,
        metadata={"help": "Maximum batch size cap."},
    )


@dataclass
class VllmSchedulerConfig(BaseReplicaSchedulerConfig):
    max_batched_tokens: int = field(
        default=None,
        metadata={"help": "Maximum batched tokens for vLLM."},
    )
    max_tokens_in_batch: int = field(
        default=4096,
        metadata={"help": "Maximum tokens in batch for vLLM."},
    )

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.VLLM


@dataclass
class LightllmSchedulerConfig(BaseReplicaSchedulerConfig):
    max_batched_tokens: int = field(
        default=None,
        metadata={"help": "Maximum batched tokens for LightLLM."},
    )
    max_tokens_in_batch: int = field(
        default=4096,
        metadata={"help": "Maximum tokens in batch for LightLLM."},
    )
    max_waiting_iters: int = field(
        default=10,
        metadata={"help": "Maximum waiting iterations for LightLLM."},
    )

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.LIGHTLLM


@dataclass
class OrcaSchedulerConfig(BaseReplicaSchedulerConfig):

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.ORCA


@dataclass
class FasterTransformerSchedulerConfig(BaseReplicaSchedulerConfig):

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.FASTER_TRANSFORMER


@dataclass
class SarathiSchedulerConfig(BaseReplicaSchedulerConfig):
    chunk_size: int = field(
        default=512,
        metadata={"help": "Chunk size for Sarathi."},
    )

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.SARATHI


@dataclass
class MnemosyneFCFSFixedChunkSchedulerConfig(BaseReplicaSchedulerConfig):
    chunk_size: int = field(
        default=512,
        metadata={"help": "Chunk size for Sarathi."},
    )

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.MNEMOSYNE_FCFS_FIXED_CHUNK


@dataclass
class MnemosyneFCFSSchedulerConfig(BaseReplicaSchedulerConfig):
    target_batch_time: float = field(
        default=0.05,
        metadata={"help": "Target batch time for Mnemosyne FCFS."},
    )

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.MNEMOSYNE_FCFS


@dataclass
class MnemosyneEDFSchedulerConfig(BaseReplicaSchedulerConfig):
    target_batch_time: float = field(
        default=0.05,
        metadata={"help": "Target batch time for Mnemosyne EDF."},
    )
    deadline_multiplier: float = field(
        default=1.5,
        metadata={"help": "Deadline multiplier for Mnemosyne EDF."},
    )
    min_deadline: float = field(
        default=0.5,
        metadata={"help": "Minimum deadline for Mnemosyne EDF."},
    )

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.MNEMOSYNE_EDF


@dataclass
class MnemosyneLRSSchedulerConfig(BaseReplicaSchedulerConfig):
    target_batch_time: float = field(
        default=0.05,
        metadata={"help": "Target batch time for Mnemosyne LRS."},
    )
    deadline_multiplier: float = field(
        default=1.5,
        metadata={"help": "Deadline multiplier for Mnemosyne LRS."},
    )
    min_deadline: float = field(
        default=0.5,
        metadata={"help": "Minimum deadline for Mnemosyne LRS."},
    )

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.MNEMOSYNE_LRS


@dataclass
class MnemosyneSTSchedulerConfig(BaseReplicaSchedulerConfig):
    target_batch_time: float = field(
        default=0.05,
        metadata={"help": "Target batch time for Mnemosyne ST."},
    )
    deadline_multiplier: float = field(
        default=1.5,
        metadata={"help": "Deadline multiplier for Mnemosyne ST."},
    )
    min_deadline: float = field(
        default=0.5,
        metadata={"help": "Minimum deadline for Mnemosyne ST."},
    )
    long_request_kv_cache_len_threshold: float = field(
        default=256 * 1024,
        metadata={
            "help": "Minimum KV cache length to be categorized as a long request."
        },
    )

    @staticmethod
    def get_type():
        return ReplicaSchedulerType.MNEMOSYNE_ST


@dataclass
class MetricsConfig:
    """Metric configuration."""

    write_metrics: bool = field(
        default=True,
        metadata={"help": "Whether to write metrics."},
    )
    write_json_trace: bool = field(
        default=False,
        metadata={"help": "Whether to write json trace."},
    )
    wandb_project: Optional[str] = field(
        default=None,
        metadata={"help": "Weights & Biases project name."},
    )
    wandb_group: Optional[str] = field(
        default=None,
        metadata={"help": "Weights & Biases group name."},
    )
    wandb_run_name: Optional[str] = field(
        default=None,
        metadata={"help": "Weights & Biases run name."},
    )
    wandb_sweep_id: Optional[str] = field(
        default=None,
        metadata={"help": "Weights & Biases sweep id."},
    )
    wandb_run_id: Optional[str] = field(
        default=None,
        metadata={"help": "Weights & Biases run id."},
    )
    enable_chrome_trace: bool = field(
        default=True,
        metadata={"help": "Enable Chrome tracing."},
    )
    save_table_to_wandb: bool = field(
        default=False,
        metadata={"help": "Whether to save table to wandb."},
    )
    store_plots: bool = field(
        default=True,
        metadata={"help": "Whether to store plots."},
    )
    store_operation_metrics: bool = field(
        default=False,
        metadata={"help": "Whether to store operation metrics."},
    )
    store_token_completion_metrics: bool = field(
        default=False,
        metadata={"help": "Whether to store token completion metrics."},
    )
    store_request_metrics: bool = field(
        default=True,
        metadata={"help": "Whether to store request metrics."},
    )
    store_batch_metrics: bool = field(
        default=True,
        metadata={"help": "Whether to store batch metrics."},
    )
    store_utilization_metrics: bool = field(
        default=True,
        metadata={"help": "Whether to store utilization metrics."},
    )
    keep_individual_batch_metrics: bool = field(
        default=False,
        metadata={"help": "Whether to keep individual batch metrics."},
    )
    subsamples: Optional[int] = field(
        default=None,
        metadata={"help": "Subsamples."},
    )
    min_batch_index: Optional[int] = field(
        default=None,
        metadata={"help": "Minimum batch index."},
    )
    max_batch_index: Optional[int] = field(
        default=None,
        metadata={"help": "Maximum batch index."},
    )
    output_dir: str = field(
        default="simulator_output",
        metadata={"help": "Output directory."},
    )

    def __post_init__(self):
        self.output_dir = (
            f"{self.output_dir}/{datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f')}"
        )
        os.makedirs(self.output_dir, exist_ok=True)


@dataclass
class ReplicaConfig:
    model_name: str = field(
        default="meta-llama/Llama-3-8B",
        metadata={"help": "Model name."},
    )
    gpu_memory_utilization: float = field(
        default=0.8,
        metadata={"help": "GPU memory utilization."},
    )
    memory_margin_fraction: float = field(
        default=0.1,
        metadata={"help": "Memory margin fraction."},
    )
    num_pipeline_stages: int = field(
        default=2,
        metadata={"help": "Number of pipeline stages."},
    )
    tensor_parallel_size: int = field(
        default=8,
        metadata={"help": "Tensor parallel size."},
    )
    enable_sequence_pipeline_parallel: bool = field(
        default=True, metadata={"help": "Enable sequence pipeline parallelism."}
    )
    kv_parallel_size: int = field(
        default=1,
        metadata={"help": "KV parallel size."},
    )
    max_num_tokens_per_kvp_group: int = field(
        default=512 * 1024,
        metadata={"help": "Maximum number of tokens per KV group."},
    )
    device: str = field(
        default="h100",
        metadata={"help": "Device."},
    )
    network_device: str = field(
        default="h100_dgx",
        metadata={"help": "Network device."},
    )
    block_size: int = field(
        default=16,
        metadata={"help": "Block size."},
    )

    def __post_init__(self):
        if self.enable_sequence_pipeline_parallel and self.num_pipeline_stages == 1:
            logger.warning(
                "Sequence pipeline parallelism is enabled but number of pipeline stages is 1."
                "Disabling sequence pipeline parallelism."
            )
            self.enable_sequence_pipeline_parallel = False

        self.world_size = self.num_pipeline_stages * self.tensor_parallel_size
        self.model_config: BaseModelConfig = BaseModelConfig.create_from_name(
            self.model_name
        )
        self.device_config: BaseDeviceSKUConfig = (
            BaseDeviceSKUConfig.create_from_type_string(self.device)
        )
        self.node_config: BaseNodeSKUConfig = BaseNodeSKUConfig.create_from_type_string(
            self.network_device
        )


@dataclass
class BaseGlobalSchedulerConfig(BasePolyConfig):
    pass


@dataclass
class RandomGlobalSchedulerConfig(BaseGlobalSchedulerConfig):
    @staticmethod
    def get_type():
        return GlobalSchedulerType.RANDOM


@dataclass
class RoundRobinGlobalSchedulerConfig(BaseGlobalSchedulerConfig):
    @staticmethod
    def get_type():
        return GlobalSchedulerType.ROUND_ROBIN


@dataclass
class LORGlobalSchedulerConfig(BaseGlobalSchedulerConfig):
    @staticmethod
    def get_type():
        return GlobalSchedulerType.LOR


@dataclass
class BaseExecutionTimePredictorConfig(BasePolyConfig):
    model_profiling_dataset: str = field(
        default="project-vajra/dev-staging-meta-llama-llama-3-8b-h100",
        metadata={
            "help": "HuggingFace dataset that contains the model (compute) profiling data."
        },
    )
    network_profiling_dataset: str = field(
        default="project-vajra/dev-staging-h100-dgx",
        metadata={
            "help": "HuggingFace dataset that contains the network (communication) profiling data."
        },
    )
    prediction_hf_org: str = field(
        default="project-vajra",
        metadata={"help": "Organization to download predictions."},
    )
    prediction_hf_collection: str = field(
        default="dev-staging",
        metadata={"help": "Collection to download predictions."},
    )
    no_cache: bool = field(
        default=False,
        metadata={"help": "Whether to cache prediction models."},
    )
    kv_cache_prediction_granularity: int = field(
        default=256,
        metadata={"help": "KV cache prediction granularity."},
    )
    prediction_max_prefill_chunk_size: int = field(
        default=4096,
        metadata={"help": "Max prefill chunk size for prediction."},
    )
    prediction_max_batch_size: int = field(
        default=128,
        metadata={"help": "Max batch size for prediction."},
    )
    prediction_max_tokens_per_request: int = field(
        default=2 * 1024 * 1024,
        metadata={"help": "Max tokens per request for prediction."},
    )
    attention_decode_batching_overhead_fraction: float = field(
        default=0.1,
        metadata={"help": "Attention decode batching overhead fraction."},
    )
    nccl_cpu_launch_overhead_ms: float = field(
        default=0.02,
        metadata={"help": "NCCL CPU launch overhead in ms."},
    )
    nccl_cpu_skew_overhead_per_device_ms: float = field(
        default=0.0,
        metadata={"help": "NCCL CPU skew overhead per device in ms."},
    )
    k_fold_cv_splits: int = field(
        default=10,
        metadata={"help": "Number of k fold cross validation splits."},
    )
    num_training_job_threads: int = field(
        default=-1,
        metadata={"help": "Number of training job threads."},
    )
    cache_dir: str = field(
        default=".vidur_cache",
        metadata={"help": "Cache directory."},
    )
    disable_kvp_communication: bool = field(
        default=True,
        metadata={"help": "Whether to disable KVP communication."},
    )
    upload_predictions: bool = field(
        default=False,
        metadata={"help": "Whether to upload predictions."},
    )
    upload_to_private_repo: bool = field(
        default=False,
        metadata={"help": "Whether to upload to private repo."},
    )
    use_native_execution_time_predictor: bool = field(
        default=True,
        metadata={"help": "Whether to use cpp execution time predictor."},
    )


@dataclass
class LinearRegressionExecutionTimePredictorConfig(BaseExecutionTimePredictorConfig):
    polynomial_degree: List[int] = field(
        default_factory=lambda: list(range(1, 6)),
        metadata={"help": "Polynomial degree for linear regression."},
    )
    polynomial_include_bias: List[bool] = field(
        default_factory=lambda: [True, False],
        metadata={"help": "Polynomial include bias for linear regression."},
    )
    polynomial_interaction_only: List[bool] = field(
        default_factory=lambda: [True, False],
        metadata={"help": "Polynomial interaction only for linear regression."},
    )
    fit_intercept: List[bool] = field(
        default_factory=lambda: [True, False],
        metadata={"help": "Fit intercept for linear regression."},
    )

    @staticmethod
    def get_type():
        return ExecutionTimePredictorType.LINEAR_REGRESSION


@dataclass
class RandomForrestExecutionTimePredictorConfig(BaseExecutionTimePredictorConfig):
    num_estimators: List[int] = field(
        default_factory=lambda: [250, 500, 750],
        metadata={"help": "Number of estimators for random forest."},
    )
    max_depth: List[int] = field(
        default_factory=lambda: [8, 16, 32],
        metadata={"help": "Maximum depth for random forest."},
    )
    min_samples_split: List[int] = field(
        default_factory=lambda: [2, 5, 10],
        metadata={"help": "Minimum samples split for random forest."},
    )

    @staticmethod
    def get_type():
        return ExecutionTimePredictorType.RANDOM_FORREST


@dataclass
class ClusterConfig:
    num_replicas: int = field(
        default=1,
        metadata={"help": "Number of replicas."},
    )
    replica_config: ReplicaConfig = field(default_factory=ReplicaConfig)
    global_scheduler_config: BaseGlobalSchedulerConfig = field(
        default_factory=RoundRobinGlobalSchedulerConfig,
        metadata={"help": "Global scheduler config."},
    )
    replica_scheduler_config: BaseReplicaSchedulerConfig = field(
        default_factory=MnemosyneFCFSFixedChunkSchedulerConfig,
        metadata={"help": "Replica scheduler config."},
    )

    def __post_init__(self):
        if self.replica_config.enable_sequence_pipeline_parallel:
            assert self.replica_scheduler_config.get_type() in [
                ReplicaSchedulerType.SARATHI,
                ReplicaSchedulerType.MNEMOSYNE_FCFS_FIXED_CHUNK,
                ReplicaSchedulerType.MNEMOSYNE_FCFS,
                ReplicaSchedulerType.MNEMOSYNE_EDF,
                ReplicaSchedulerType.MNEMOSYNE_LRS,
                ReplicaSchedulerType.MNEMOSYNE_ST,
            ]


@dataclass
class SimulationConfig(ABC):
    seed: int = field(
        default=42,
        metadata={"help": "Seed for the random number generator."},
    )
    log_level: str = field(
        default="info",
        metadata={"help": "Logging level."},
    )
    time_limit: int = field(
        default=0,  # in seconds, 0 is no limit
        metadata={"help": "Time limit for simulation in seconds. 0 means no limit."},
    )
    cluster_config: ClusterConfig = field(
        default_factory=ClusterConfig,
        metadata={"help": "Cluster config."},
    )
    request_generator_config: BaseRequestGeneratorConfig = field(
        default_factory=SyntheticRequestGeneratorConfig,
        metadata={"help": "Request generator config."},
    )
    execution_time_predictor_config: BaseExecutionTimePredictorConfig = field(
        default_factory=RandomForrestExecutionTimePredictorConfig,
        metadata={"help": "Execution time predictor config."},
    )
    metrics_config: MetricsConfig = field(
        default_factory=MetricsConfig,
        metadata={"help": "Metrics config."},
    )

    def __post_init__(self):
        self.write_config_to_file()

    @classmethod
    def create_from_cli_args(cls):
        flat_config = create_flat_dataclass(cls).create_from_cli_args()
        instance = flat_config.reconstruct_original_dataclass()
        instance.__flat_config__ = flat_config
        return instance

    def to_dict(self):
        if not hasattr(self, "__flat_config__"):
            logger.warning("Flat config not found. Returning the original config.")
            return self.__dict__

        return self.__flat_config__.__dict__

    def write_config_to_file(self):
        config_dict = dataclass_to_dict(self)
        with open(f"{self.metrics_config.output_dir}/config.json", "w") as f:
            json.dump(config_dict, f, indent=4)



from abc import ABC, abstractmethod

import numpy as np

TTFT_THRESHOLD = 0.8
TBT_THRESHOLD = 0.8
LOW_BATCH_SIZE_PERCENTILE = 75
LOW_BATCH_SIZE_THRESHOLD = 0.9
LOW_MEMORY_MEMORY_THRESHOLD = 0.7
LOW_MEMORY_BATCH_SIZE_PERCENTILE = 75
LOW_MEMORY_BATCH_SIZE_THRESHOLD = 0.7
LOW_PREFILL_THROUGHPUT_PIPELINE_BUBBLES_IDLE_THRESHOLD = 0.4
HIGH_MODEL_EXECUTION_LATENCY_PERCENTILE = 25
HIGH_MODEL_EXECUTION_LATENCY_PERCENTILE_THRESHOLD = 0.75


str_to_list = lambda x: list(map(float, x[1:-1].split(",")))


class BaseBottleneckCase(ABC):
    def __init__(
        self,
        config_row,
        ttft_slo_percentile: float,
        ttft_slo_value: float,
        tbt_slo_percentile: float,
        tbt_slo_value: float,
    ) -> None:
        self.config_row = config_row
        self.ttft_slo_percentile = ttft_slo_percentile
        self.ttft_slo_value = ttft_slo_value
        self.tbt_slo_percentile = tbt_slo_percentile
        self.tbt_slo_value = tbt_slo_value

    @abstractmethod
    def is_match(self) -> bool:
        return True

    @abstractmethod
    def get_message(self) -> str:
        return ""


class TTFTViolationCase(BaseBottleneckCase):
    def is_match(self):
        super_is_match = super().is_match()
        return (
            super_is_match
            and self.config_row[f"TTFT_{self.ttft_slo_percentile}%"]
            > self.ttft_slo_value * TTFT_THRESHOLD
        )

    def get_message(self):
        super_get_message = super().get_message()
        return f"{super_get_message} TTFT (P{self.ttft_slo_percentile}) SLO violation risk."


class TBTViolationCase(BaseBottleneckCase):
    def is_match(self):
        super_is_match = super().is_match()
        return (
            super_is_match
            and self.config_row[f"TBT_{self.tbt_slo_percentile}%"]
            > self.tbt_slo_value * TBT_THRESHOLD
        )

    def get_message(self):
        super_get_message = super().get_message()
        return (
            f"{super_get_message} TBT (P{self.tbt_slo_percentile}) SLO violation risk."
        )


class TFFTViolationLowMaxBatchSizeCase(TTFTViolationCase):
    def is_match(self):
        super_is_match = super().is_match()
        # check if P{X} batch size greater than {Y} * max batch size
        batch_size_cdf = str_to_list(self.config_row["batch_size_cdf"])

        batch_size_obs = np.percentile(batch_size_cdf, LOW_BATCH_SIZE_PERCENTILE)

        max_batch_size = self.config_row["Batch Size"]

        return (
            super_is_match
            and batch_size_obs > LOW_BATCH_SIZE_THRESHOLD * max_batch_size
        )

    def get_message(self):
        super_get_message = super().get_message()
        return f"{super_get_message} Increase max batch size."


class TFFTViolationLowMemoryCase(TTFTViolationCase):
    def is_match(self):
        super_is_match = super().is_match()
        # check if memory usage is high
        memory_usage = self.config_row["memory_usage_mean"]
        mean_memory_usage = np.mean(memory_usage)
        # and batch size is lower than limit
        batch_size_cdf = str_to_list(self.config_row["batch_size_cdf"])

        batch_size_obs = np.percentile(batch_size_cdf, LOW_MEMORY_BATCH_SIZE_PERCENTILE)

        return (
            super_is_match
            and mean_memory_usage > LOW_MEMORY_MEMORY_THRESHOLD
            and batch_size_obs < LOW_MEMORY_BATCH_SIZE_THRESHOLD
        )

    def get_message(self):
        super_get_message = super().get_message()
        return (
            f"{super_get_message} Low memory availability, increase parallelism degree."
        )


class TFFTViolationLowPrefillThroughputCase(TTFTViolationCase):
    def is_match(self):
        super_is_match = super().is_match()
        # just check if none of the other cases apply
        is_low_max_batch_size = TFFTViolationLowMaxBatchSizeCase(
            self.config_row,
            self.ttft_slo_percentile,
            self.ttft_slo_value,
            self.tbt_slo_percentile,
            self.tbt_slo_value,
        ).is_match()
        is_low_memory = TFFTViolationLowMemoryCase(
            self.config_row,
            self.ttft_slo_percentile,
            self.ttft_slo_value,
            self.tbt_slo_percentile,
            self.tbt_slo_value,
        ).is_match()

        return super_is_match and not is_low_max_batch_size and not is_low_memory

    def get_message(self):
        super_get_message = super().get_message()
        return f"{super_get_message} Low prefill throughput."


class PipelineBubblesCase(BaseBottleneckCase):
    def is_match(self):
        super_is_match = super().is_match()
        # check if low efficiency due to pipeline bubbles
        idle_time = 100 - self.config_row["busy_time_percent_mean"]
        return (
            super_is_match
            and idle_time > LOW_PREFILL_THROUGHPUT_PIPELINE_BUBBLES_IDLE_THRESHOLD * 100
        )

    def get_message(self):
        super_get_message = super().get_message()
        return f"{super_get_message} High pipeline bubbles causing low performance."


class TFFTViolationLowPrefillThroughputPipelineBubblesCase(
    PipelineBubblesCase, TFFTViolationLowPrefillThroughputCase
):
    pass


class TBTViolationHighModelExecutionLatencyCase(TBTViolationCase):
    def is_match(self):
        super_is_match = super().is_match()
        low_percentile_tbt = self.config_row[
            f"TBT_{HIGH_MODEL_EXECUTION_LATENCY_PERCENTILE}%"
        ]
        return (
            super_is_match
            and low_percentile_tbt
            > self.tbt_slo_value * HIGH_MODEL_EXECUTION_LATENCY_PERCENTILE_THRESHOLD
        )

    def get_message(self):
        super_get_message = super().get_message()
        return f"{super_get_message} High model execution latency, increase parallelism degree."


class TBTViolationHighModelExecutionLatencyPipelineBubblesCase(
    PipelineBubblesCase, TBTViolationHighModelExecutionLatencyCase
):
    pass


class TBTViolationHighTailLatencyCase(TBTViolationCase):
    def is_match(self):
        super_is_match = super().is_match()
        low_percentile_tbt = self.config_row[
            f"TBT_{HIGH_MODEL_EXECUTION_LATENCY_PERCENTILE}%"
        ]
        tail_tbt = self.config_row[f"TBT_{self.tbt_slo_percentile}%"]
        return (
            super_is_match
            and low_percentile_tbt
            < self.tbt_slo_value * HIGH_MODEL_EXECUTION_LATENCY_PERCENTILE_THRESHOLD
            and tail_tbt > self.tbt_slo_value * TBT_THRESHOLD
        )

    def get_message(self):
        super_get_message = super().get_message()
        return f"{super_get_message} High tail latency."


class TBTViolationHighTailLatencySchedulerCase(TBTViolationHighTailLatencyCase):
    def is_match(self):
        super_is_match = super().is_match()
        return super_is_match and self.config_row["Scheduler"] != "Sarathi-Serve"

    def get_message(self):
        super_get_message = super().get_message()
        return f"{super_get_message} Switch to Sarathi-Serve scheduler to avoid prefill-decode interference."


class TBTViolationHighTailLatencySarathiChunkSizeCase(TBTViolationHighTailLatencyCase):
    def is_match(self):
        super_is_match = super().is_match()
        return (
            super_is_match
            and self.config_row["Scheduler"] == "Sarathi-Serve"
            and self.config_row["Sarathi Chunk Size"] > 1024
        )

    def get_message(self):
        super_get_message = super().get_message()
        return f"{super_get_message} Reduce chunk size."



import numpy as np
import pandas as pd
import plotly_express as px
import wandb
from ddsketch.ddsketch import DDSketch

from vidur.logger import init_logger

logger = init_logger(__name__)


class CDFSketch:
    def __init__(
        self,
        metric_name: str,
        save_table_to_wandb: bool = True,
        save_plots: bool = True,
    ) -> None:
        # metrics are a data series of two-dimensional (x, y) datapoints
        self._sketch = DDSketch(relative_accuracy=0.001)
        # column name
        self._metric_name = metric_name

        # most recently collected y datapoint for incremental updates
        # to aid incremental updates to y datapoints
        self._last_data = 0

        self._save_table_to_wandb = save_table_to_wandb
        self._save_plots = save_plots

    def __len__(self):
        return int(self._sketch.count)

    # add a new x, y datapoint
    def put(self, data: float) -> None:
        self._last_data = data
        self._sketch.add(data)

    # add a new datapoint as an incremental (delta) update to
    # recently collected datapoint
    def put_delta(self, delta: float) -> None:
        data = self._last_data + delta
        self.put(data)

    def print_distribution_stats(self, plot_name: str) -> None:
        if self._sketch._count == 0:
            return

        logger.debug(
            f"{plot_name}: {self._metric_name} stats:"
            f" min: {self._sketch._min},"
            f" max: {self._sketch._max},"
            f" mean: {self._sketch.avg},"
            f" 25th percentile: {self._sketch.get_quantile_value(0.25)},"
            f" median: {self._sketch.get_quantile_value(0.5)},"
            f" 75th percentile: {self._sketch.get_quantile_value(0.75)},"
            f" 95th percentile: {self._sketch.get_quantile_value(0.95)},"
            f" 99th percentile: {self._sketch.get_quantile_value(0.99)}"
            f" 99.9th percentile: {self._sketch.get_quantile_value(0.999)}"
            f" count: {self._sketch._count}"
            f" sum: {self._sketch.sum}"
        )
        if wandb.run:
            wandb.log(
                {
                    f"{plot_name}_min": self._sketch._min,
                    f"{plot_name}_max": self._sketch._max,
                    f"{plot_name}_mean": self._sketch.avg,
                    f"{plot_name}_25th_percentile": self._sketch.get_quantile_value(
                        0.25
                    ),
                    f"{plot_name}_median": self._sketch.get_quantile_value(0.5),
                    f"{plot_name}_75th_percentile": self._sketch.get_quantile_value(
                        0.75
                    ),
                    f"{plot_name}_95th_percentile": self._sketch.get_quantile_value(
                        0.95
                    ),
                    f"{plot_name}_99th_percentile": self._sketch.get_quantile_value(
                        0.99
                    ),
                    f"{plot_name}_99.9th_percentile": self._sketch.get_quantile_value(
                        0.999
                    ),
                    f"{plot_name}_count": self._sketch.count,
                    f"{plot_name}_sum": self._sketch.sum,
                },
                step=0,
            )

    def _to_df(self) -> pd.DataFrame:
        # get quantiles at 1% intervals
        quantiles = np.linspace(0, 1, 101)
        # get quantile values
        quantile_values = [self._sketch.get_quantile_value(q) for q in quantiles]
        # create dataframe
        df = pd.DataFrame({"cdf": quantiles, self._metric_name: quantile_values})

        return df

    @property
    def sum(self) -> float:
        return self._sketch.sum

    def _save_df(self, df: pd.DataFrame, path: str, plot_name: str) -> None:
        df.to_csv(f"{path}/{plot_name}.csv")

        if wandb.run and self._save_table_to_wandb:
            wand_table = wandb.Table(dataframe=df)
            wandb.log({f"{plot_name}_table": wand_table}, step=0)

    def plot_cdf(self, path: str, plot_name: str, x_axis_label: str = None) -> None:
        if self._sketch._count == 0:
            return

        if x_axis_label is None:
            x_axis_label = self._metric_name

        df = self._to_df()

        self.print_distribution_stats(plot_name)

        if wandb.run:
            wandb_df = df.copy()
            # rename the self._metric_name column to x_axis_label
            wandb_df = wandb_df.rename(columns={self._metric_name: x_axis_label})

            wandb.log(
                {
                    f"{plot_name}_cdf": wandb.plot.line(
                        wandb.Table(dataframe=wandb_df),
                        "cdf",
                        x_axis_label,
                        title=plot_name,
                    )
                },
                step=0,
            )

        if self._save_plots:
            fig = px.line(
                df,
                x=self._metric_name,
                y="cdf",
                markers=True,
                labels={"x": x_axis_label},
            )
            fig.update_traces(marker=dict(color="red", size=2))
            fig.write_image(f"{path}/{plot_name}.png")
        self._save_df(df, path, plot_name)



from dataclasses import dataclass, field

from vidur.config.base_fixed_config import BaseFixedConfig
from vidur.logger import init_logger
from vidur.types import DeviceSKUType

logger = init_logger(__name__)


@dataclass
class BaseDeviceSKUConfig(BaseFixedConfig):
    fp16_tflops: int
    total_memory_gb: int


@dataclass
class A40DeviceSKUConfig(BaseDeviceSKUConfig):
    fp16_tflops: int = 150
    total_memory_gb: int = 45

    @staticmethod
    def get_type():
        return DeviceSKUType.A40


@dataclass
class A100DeviceSKUConfig(BaseDeviceSKUConfig):
    fp16_tflops: int = 312
    total_memory_gb: int = 80

    @staticmethod
    def get_type():
        return DeviceSKUType.A100


@dataclass
class H100DeviceSKUConfig(BaseDeviceSKUConfig):
    fp16_tflops: int = 1000
    total_memory_gb: int = 80

    @staticmethod
    def get_type():
        return DeviceSKUType.H100



import plotly.express as px
import streamlit as st

from vidur.config_optimizer.analyzer.bottleneck_analyzer import BottleneckAnalyzer
from vidur.config_optimizer.analyzer.constants import AXIS_COLS
from vidur.config_optimizer.analyzer.dashboard.utils import (
    add_advanced_filters,
    add_model_trace_selector,
    add_small_divider,
    convert_config_row_to_comparison_point,
    write_best_config,
)


def render_axis_comparison_bar_chart(
    df, metric_1, percentile_1, slo_1, metric_2, percentile_2, slo_2, axis_col
):
    if axis_col == "Sarathi Chunk Size":
        df = df[df["Scheduler"] == "Sarathi-Serve"]

    best_configs_df = (
        df[
            (df[f"{metric_1}_{percentile_1}%"] <= slo_1)
            & (df[f"{metric_2}_{percentile_2}%"] <= slo_2)
        ]
        .sort_values("QPS", ascending=False)
        .groupby(axis_col)
        .first()
        .reset_index()
    )

    # treat each x axis as a category
    best_configs_df[f"{axis_col}_str"] = best_configs_df[axis_col].astype(str)

    fig = px.bar(
        best_configs_df,
        x=f"{axis_col}_str",
        y="QPS",
        color=f"{axis_col}_str",
        hover_data=list(AXIS_COLS.values()),
        labels={f"{axis_col}_str": f"{axis_col}"},
        width=300,
        height=300,
    )

    fig.update_xaxes(type="category")

    st.plotly_chart(fig)

    add_small_divider()

    add_to_comparison_button = st.button(
        f"Add to Comparison", key=f"{axis_col}_add_to_comparison"
    )

    if add_to_comparison_button:
        points_to_add = best_configs_df.apply(
            convert_config_row_to_comparison_point, axis=1
        ).to_list()
        for point in points_to_add:
            if point not in st.session_state.comparison_points:
                st.session_state.comparison_points.append(point)


def render_cost_analysis(
    bottleneck_analyzer,
    df,
    metric_1,
    percentile_1,
    slo_1,
    metric_2,
    percentile_2,
    slo_2,
    cost_budget,
):
    num_replicas = cost_budget // df["hour_cost_per_replica"]
    capacity_per_replica = (
        df["poisson_request_interval_generator_qps"] / df["cluster_num_replicas"]
    )
    df["QPS"] = num_replicas * capacity_per_replica

    write_best_config(
        bottleneck_analyzer,
        df,
        metric_1,
        percentile_1,
        slo_1,
        metric_2,
        percentile_2,
        slo_2,
        use_qps=True,
    )
    cols = st.columns(3)

    for i, axis_col in enumerate(AXIS_COLS.values()):
        with cols[i % 3]:
            with st.container(border=True):
                render_axis_comparison_bar_chart(
                    df,
                    metric_1,
                    percentile_1,
                    slo_1,
                    metric_2,
                    percentile_2,
                    slo_2,
                    axis_col,
                )


def render_cost_analysis_page(subset_dfs):
    # create a subpage for best config selection
    st.markdown("## Cost Analysis")
    st.markdown(
        "This tool helps you select the best configuration based on your SLOs and cost budget. It also provides additional visualizations which allow to see the performance as a function of different config options."
    )

    add_small_divider()
    st.markdown("### Input Parameters")

    model_select_box, trace_select_box = add_model_trace_selector(subset_dfs)

    subset_df = subset_dfs[(model_select_box, trace_select_box)]

    col1, col2 = st.columns(2)

    with col1:
        ttft_slo_percentile = st.selectbox(
            "TTFT SLO Percentile:", [50, 75, 90, 95, 99], index=2
        )
        ttft_slo_value = st.slider(
            "TTFT SLO Value (s)", min_value=0.1, max_value=10.0, value=2.0
        )

    with col2:
        tbt_slo_percentile = st.selectbox(
            "TBT SLO Percentile:", [50, 75, 90, 95, 99], index=4
        )
        tbt_slo_value = st.slider(
            "TBT SLO Value (ms)", min_value=10, max_value=2000, value=200
        )

    cost_budget = st.slider(
        "Per Hour Cost Budget ($)", min_value=0, max_value=1000, value=100
    )

    filtered_df = add_advanced_filters(subset_df)

    bottleneck_analyzer = BottleneckAnalyzer(
        ttft_slo_percentile,
        ttft_slo_value,
        tbt_slo_percentile,
        tbt_slo_value,
    )

    st.markdown("### Results")
    add_small_divider()

    render_cost_analysis(
        bottleneck_analyzer,
        filtered_df,
        "TTFT",
        ttft_slo_percentile,
        ttft_slo_value,
        "TBT",
        tbt_slo_percentile,
        tbt_slo_value,
        cost_budget,
    )



from typing import List

from vidur.entities.base_entity import BaseEntity
from vidur.entities.execution_time import ExecutionTime
from vidur.entities.request import Request
from vidur.logger import init_logger

logger = init_logger(__name__)


# a decorator which checks if the request has been scheduled
def check_scheduled(func):
    def wrapper(self, *args, **kwargs):
        if not self._scheduled:
            raise ValueError("Batch has not been scheduled yet")
        return func(self, *args, **kwargs)

    return wrapper


class BatchStage(BaseEntity):
    def __init__(
        self,
        batch_id: int,
        replica_id: int,
        pipeline_stage: int,
        execution_time: ExecutionTime,
        requests: List[Request],
        num_q_tokens: List[int],
        num_kv_tokens: List[int],
        num_active_kvp_groups: List[int] = None,
        last_kvp_group_ids: List[int] = None,
        kvp_group_id: int = 0,
    ) -> None:
        self._id = BatchStage.generate_id()

        self._requests = requests
        self._num_q_tokens = num_q_tokens
        self._num_kv_tokens = num_kv_tokens
        self._num_active_kvp_groups = num_active_kvp_groups
        self._last_kvp_group_ids = last_kvp_group_ids
        self._batch_id = batch_id
        self._replica_id = replica_id
        self._pipeline_stage = pipeline_stage
        self._execution_time = execution_time
        self._kvp_group_id = kvp_group_id

        self._total_execution_time = self._execution_time.total_time
        self._model_execution_time = self._execution_time.model_time

        self._scheduled_at = None
        self._completed_at = None
        self._scheduled = False

    @property
    def num_q_tokens(self) -> List[int]:
        return self._num_q_tokens

    @property
    def num_kv_tokens(self) -> List[int]:
        return self._num_kv_tokens

    @property
    @check_scheduled
    def scheduled_at(self) -> float:
        return self._scheduled_at

    @property
    @check_scheduled
    def completed_at(self) -> float:
        return self._completed_at

    @property
    def execution_time(self) -> float:
        return self._total_execution_time

    @property
    def model_execution_time(self) -> float:
        return self._model_execution_time

    @property
    def pipeline_stage(self) -> int:
        return self._pipeline_stage

    @property
    def request_ids(self) -> List[int]:
        return [request.id for request in self._requests]

    @property
    def requests(self) -> List[Request]:
        return self._requests

    @property
    def size(self) -> int:
        return len(self._requests)

    def on_schedule(
        self,
        time: float,
    ) -> None:
        self._scheduled_at = time
        self._scheduled = True

        if not self._last_kvp_group_ids:
            for request in self._requests:
                request.on_batch_stage_schedule(time)
            return

        for request, last_kvp_group_id in zip(self._requests, self._last_kvp_group_ids):
            if self._kvp_group_id != last_kvp_group_id:
                continue
            request.on_batch_stage_schedule(time)

    def on_stage_end(
        self,
        time: float,
    ) -> None:
        assert (
            time == self._scheduled_at + self._total_execution_time
        ), f"{time} != {self._scheduled_at} + {self._total_execution_time}"

        self._completed_at = time

        if not self._last_kvp_group_ids:
            for request, num_q_tokens in zip(self._requests, self._num_q_tokens):
                request.on_batch_stage_end(
                    time,
                    self._pipeline_stage,
                    self._total_execution_time,
                    self._model_execution_time,
                    num_q_tokens,
                )
            return

        for request, num_q_tokens, last_kvp_group_id in zip(
            self._requests, self._num_q_tokens, self._last_kvp_group_ids
        ):
            if self._kvp_group_id != last_kvp_group_id:
                continue

            request.on_batch_stage_end(
                time,
                self._pipeline_stage,
                self._total_execution_time,
                self._model_execution_time,
                num_q_tokens,
            )

    def to_dict(self) -> dict:
        return {
            "id": self._id,
            "size": self.size,
            "execution_time": self._execution_time.to_dict(),
            "scheduled_at": self._scheduled_at,
            "completed_at": self._completed_at,
            "replica_id": self._replica_id,
            "batch_id": self._batch_id,
            "pipeline_stage": self._pipeline_stage,
            "scheduled": self._scheduled,
            "request_ids": self.request_ids,
            "num_q_tokens": self.num_q_tokens,
            "num_kv_tokens": self.num_kv_tokens,
        }

    def to_chrome_trace_events(self, time: int) -> dict:
        return [
            {
                "name": f"{self.request_ids}",
                "ph": "X",
                "ts": (time - self._total_execution_time) * 1e6,
                "dur": self._total_execution_time * 1e6,
                "pid": f"replica_{self._replica_id}_kvp_group_{self._kvp_group_id}",
                "tid": self._pipeline_stage,
                "args": {
                    "batch_id": self._batch_id,
                    "batch_size": self.size,
                    "request_ids": self.request_ids,
                    "num_q_tokens": self.num_q_tokens,
                    "num_kv_tokens": self.num_kv_tokens,
                    # "execution_time": self._execution_time.to_dict(),
                    # "requests": [request.to_dict() for request in self._requests],  # slow sim time, commented out for now
                },
            }
        ]



from abc import ABC, abstractmethod
from typing import Union

from vidur._native.entities import Batch as BatchC
from vidur._native.entities import ExecutionTime as ExecutionTimeC
from vidur._native.entities import KVParallelBatch as KVParallelBatchC
from vidur.config import BaseExecutionTimePredictorConfig, ReplicaConfig
from vidur.entities import Batch, KVParallelBatch


class BaseExecutionTimePredictor(ABC):
    def __init__(
        self,
        predictor_config: BaseExecutionTimePredictorConfig,
        replica_config: ReplicaConfig,
    ) -> None:
        self._config = predictor_config
        self._replica_config = replica_config
        self._model_config = replica_config.model_config

        # get configs
        self._block_size = replica_config.block_size
        self._cache_dir = predictor_config.cache_dir
        self._num_layers_per_pipeline_stage = (
            self._model_config.num_layers // self._replica_config.num_pipeline_stages
        )
        self._native_execution_time_predictor = None

    def get_execution_time_native(
        self, batch: Union[Batch, KVParallelBatch], pipeline_stage: int
    ) -> ExecutionTimeC:
        if isinstance(batch, KVParallelBatch):
            kvp_group_ids, batches = zip(*batch.batch_mapping.items())

            batches_native = [
                BatchC(
                    replica_id=batch.replica_id,
                    num_requests=len(batch.requests),
                    num_q_tokens=batch.num_q_tokens,
                    num_kv_tokens=batch.num_kv_tokens,
                    num_active_kvp_groups=batch.num_active_kvp_groups,
                    kvp_group_id=batch.kvp_group_id,
                )
                for batch in batches
            ]

            kv_parallel_batch = KVParallelBatchC(
                replica_id=batch.replica_id,
                kvp_group_ids=kvp_group_ids,
                batches=batches_native,
            )
            return self._native_execution_time_predictor.get_execution_time_kv_parallel_batch(
                kv_parallel_batch, pipeline_stage
            )

        batch = BatchC(
            replica_id=batch.replica_id,
            num_requests=len(batch.requests),
            num_q_tokens=batch.num_q_tokens,
            num_kv_tokens=batch.num_kv_tokens,
            num_active_kvp_groups=batch.num_active_kvp_groups,
            kvp_group_id=batch.kvp_group_id,
        )
        return self._native_execution_time_predictor.get_execution_time_batch(
            batch, pipeline_stage
        )

    def get_execution_time(
        self, batch: Union[Batch, KVParallelBatch], pipeline_stage: int
    ) -> ExecutionTimeC:
        if self._config.use_native_execution_time_predictor:
            return self.get_execution_time_native(batch, pipeline_stage)

        if isinstance(batch, KVParallelBatch):
            # take the max of all the kv group execution times
            kvp_group_execution_time = max(
                [
                    self.get_execution_time(sub_batch, pipeline_stage)
                    for sub_batch in batch.batch_mapping.values()
                ],
                key=lambda x: x.total_time,
            )
            return kvp_group_execution_time

        if pipeline_stage == self._replica_config.num_pipeline_stages - 1:
            pipeline_parallel_communication_time = 0
        else:
            pipeline_parallel_communication_time = (
                self._get_pipeline_parallel_communication_time(batch)
            )

        if self._replica_config.tensor_parallel_size == 1:
            tensor_parallel_communication_time = 0
        else:
            tensor_parallel_communication_time = (
                self._get_tensor_parallel_communication_time(batch)
            )

        # TODO(Amey): We aren't adding the kvp communication time yet
        # due to some missing data points. Even though rest of the modeling
        # code is already in place.

        return ExecutionTimeC(
            self._num_layers_per_pipeline_stage,
            self._get_attention_rope_execution_time(batch),
            self._get_attention_kv_cache_save_execution_time(batch),
            self._get_attention_decode_execution_time(batch),
            self._get_attention_prefill_execution_time(batch),
            self._get_attention_layer_pre_proj_execution_time(batch),
            self._get_attention_layer_post_proj_execution_time(batch),
            self._get_mlp_layer_up_proj_execution_time(batch),
            self._get_mlp_layer_down_proj_execution_time(batch),
            self._get_mlp_layer_act_execution_time(batch),
            self._get_attn_norm_layer_act_execution_time(batch),
            self._get_mlp_norm_layer_act_execution_time(batch),
            self._get_add_layer_act_execution_time(batch),
            tensor_parallel_communication_time,
            pipeline_parallel_communication_time,
        )

    @abstractmethod
    def _get_attention_layer_pre_proj_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_attention_layer_post_proj_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_attention_rope_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_attention_kv_cache_save_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_attention_decode_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_attention_prefill_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_mlp_layer_up_proj_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_mlp_layer_down_proj_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_mlp_layer_act_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_tensor_parallel_communication_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_pipeline_parallel_communication_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_mlp_norm_layer_act_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_attn_norm_layer_act_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_add_layer_act_execution_time(self, batch: Batch) -> float:
        pass

    @abstractmethod
    def _get_kv_parallel_communication_time(self, batch: Batch) -> float:
        pass
